import warnings
from itertools import product as iproduct
from typing import Optional

import numpy as np
from scipy.special import binom
from scipy.special import perm
from sklearn.utils.validation import check_is_fitted

from ..utils import AxesArray
from ..utils import comprehend_axes
from .base import BaseFeatureLibrary
from .base import x_sequence_or_item
from .polynomial_library import PolynomialLibrary
from pysindy.differentiation import FiniteDifference


class WeakPDELibrary(BaseFeatureLibrary):
    """Generate a weak formulation library with custom functions and,
       optionally, any spatial derivatives in arbitrary dimensions.

       The features in the weak formulation are integrals of derivatives of input data
       multiplied by a test function phi, which are evaluated on K subdomains
       randomly sampled across the spatiotemporal grid. Each subdomain
       is initial generated with a size H_xt along each axis, and is then shrunk
       such that the left and right boundaries lie on spatiotemporal grid points.
       The expressions are integrated by parts to remove as many derivatives from the
       input data as possible and put the derivatives onto the test functions.

       The weak integral features are calculated assuming the function f(x) to
       integrate against derivatives of the test function dphi(x)
       is linear between grid points provided by the data:
       f(x)=f_i+(x-x_i)/(x_{i+1}-x_i)*(f_{i+1}-f_i)
       Thus f(x)*dphi(x) is approximated as a piecewise polynomial.
       The piecewise components are integrated analytically. To improve performance,
       the complete integral is expressed as a dot product of weights against the
       input data f_i, which enables vectorized evaulations.

    Parameters
    ----------
    function_library : BaseFeatureLibrary, optional (default
        PolynomialLibrary(degree=3,include_bias=False))
        SINDy library with output features representing library_functions to include
        in the library, in place of library_functions.

    derivative_order : int, optional (default 0)
        Order of derivative to take on each input variable,
        can be arbitrary non-negative integer.

    spatiotemporal_grid : np.ndarray (default None)
        The spatiotemporal grid for computing derivatives.
        This variable must be specified with
        at least one dimension corresponding to a temporal grid, so that
        integration by parts can be done in the weak formulation.

    include_bias : boolean, optional (default False)
        If True (default), then include a bias column, the feature in which
        all polynomial powers are zero (i.e. a column of ones - acts as an
        intercept term in a linear model).
        This is hard to do with just lambda functions, because if the system
        is not 1D, lambdas will generate duplicates.

    include_interaction : boolean, optional (default True)
        This is a different than the use for the PolynomialLibrary. If true,
        it generates all the mixed derivative terms. If false, the library
        will consist of only pure no-derivative terms and pure derivative
        terms, with no mixed terms.

    K : int, optional (default 100)
        Number of domain centers, corresponding to subdomain squares of length
        Hxt. If K is not
        specified, defaults to 100.

    H_xt : array of floats, optional (default None)
        Half of the length of the square subdomains in each spatiotemporal
        direction. If H_xt is not specified, defaults to H_xt = L_xt / 20,
        where L_xt is the length of the full domain in each spatiotemporal
        direction. If H_xt is specified as a scalar, this value will be applied
        to all dimensions of the subdomains.

    p : int, optional (default 4)
        Positive integer to define the polynomial degree of the spatial weights
        used for weak/integral SINDy.

    num_pts_per_domain : int, deprecated (default None)
        Included here to retain backwards compatibility with older code
        that uses this parameter. However, it merely raises a
        DeprecationWarning and then is ignored.

    implicit_terms : boolean
        Flag to indicate if SINDy-PI (temporal derivatives) is being used
        for the right-hand side of the SINDy fit.

    multiindices : list of integer arrays,  (default None)
        Overrides the derivative_order to customize the included derivative
        orders. Each integer array indicates the order of differentiation
        along the corresponding axis for each derivative term.

    differentiation_method : callable,  (default FiniteDifference)
        Spatial differentiation method.

     diff_kwargs: dictionary,  (default {})
        Keyword options to supply to differtiantion_method.


    Attributes
    ----------
    n_features_in_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. The number of output features
        is the product of the number of library functions and the number of
        input features.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.feature_library import WeakPDELibrary
    >>> x = np.array([[0.,-1],[1.,0.],[2.,-1.]])
    >>> functions = [lambda x : np.exp(x), lambda x,y : np.sin(x+y)]
    >>> lib = WeakPDELibrary(library_functions=functions).fit(x)
    >>> lib.transform(x)
    array([[ 1.        ,  0.36787944, -0.84147098],
           [ 2.71828183,  1.        ,  0.84147098],
           [ 7.3890561 ,  0.36787944,  0.84147098]])
    >>> lib.get_feature_names()
    ['f0(x0)', 'f0(x1)', 'f1(x0,x1)']
    """

    def __init__(
        self,
        function_library: Optional[BaseFeatureLibrary] = None,
        derivative_order=0,
        spatiotemporal_grid=None,
        include_bias=False,
        include_interaction=True,
        K=100,
        H_xt=None,
        p=4,
        num_pts_per_domain=None,
        implicit_terms=False,
        multiindices=None,
        differentiation_method=FiniteDifference,
        diff_kwargs={},
        is_uniform=None,
        periodic=None,
    ):
        self.function_library = function_library
        self.derivative_order = derivative_order
        self.implicit_terms = implicit_terms
        self.include_bias = include_bias
        self.include_interaction = include_interaction
        self.K = K
        self.H_xt = H_xt
        self.p = p
        self.num_trajectories = 1
        self.differentiation_method = differentiation_method
        self.diff_kwargs = diff_kwargs
        self.multiindices = multiindices
        self.spatiotemporal_grid = spatiotemporal_grid
        if self.function_library is None:
            self.function_library = PolynomialLibrary(degree=3, include_bias=False)

        if spatiotemporal_grid is None:
            raise ValueError(
                "Spatiotemporal grid was not passed, and at least a 1D"
                " grid is required, corresponding to the time base."
            )
        if num_pts_per_domain is not None:
            warnings.warn(
                "The parameter num_pts_per_domain is now deprecated. This "
                "value will be ignored by the library."
            )
        if is_uniform is not None or periodic is not None:
            # DeprecationWarning are ignored by default...
            warnings.warn(
                "is_uniform and periodic have been deprecated."
                "in favor of differetiation_method and diff_kwargs.",
                UserWarning,
            )
        # Weak form checks and setup
        self._weak_form_setup()

    def set_params(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        self._weak_form_setup()

    def _weak_form_setup(self):
        # list of integrals
        indices = ()
        if np.array(self.spatiotemporal_grid).ndim == 1:
            self.spatiotemporal_grid = np.reshape(
                self.spatiotemporal_grid, (len(self.spatiotemporal_grid), 1)
            )
        dims = self.spatiotemporal_grid.shape[:-1]
        self.grid_dims = dims
        self.grid_ndim = len(dims)

        # if want to include temporal terms -> range(len(dims))
        if self.implicit_terms:
            self.ind_range = len(dims)
        else:
            self.ind_range = len(dims) - 1

        for i in range(self.ind_range):
            indices = indices + (range(self.derivative_order + 1),)

        if self.multiindices is None:
            multiindices = []
            for ind in iproduct(*indices):
                current = np.array(ind)
                if np.sum(ind) > 0 and np.sum(ind) <= self.derivative_order:
                    multiindices.append(current)
            self.multiindices = np.array(multiindices)
        num_derivatives = len(self.multiindices)
        if num_derivatives > 0:
            self.derivative_order = np.max(self.multiindices)

        self.num_derivatives = num_derivatives
        self.multiindices = multiindices

        self.spatiotemporal_grid = AxesArray(
            self.spatiotemporal_grid, axes=comprehend_axes(self.spatiotemporal_grid)
        )
        xt1, xt2 = self._get_spatial_endpoints()
        L_xt = xt2 - xt1
        if self.H_xt is not None:
            if np.isscalar(self.H_xt):
                self.H_xt = AxesArray(
                    np.array(self.grid_ndim * [self.H_xt]), {"ax_coord": 0}
                )
            if self.grid_ndim != len(self.H_xt):
                raise ValueError(
                    "The user-defined grid (spatiotemporal_grid) and "
                    "the user-defined sizes of the subdomains for the "
                    "weak form do not have the same # of spatiotemporal "
                    "dimensions. For instance, if spatiotemporal_grid is 4D, "
                    "then H_xt should be a 4D list of the subdomain lengths."
                )
            if any(self.H_xt <= np.zeros(len(self.H_xt))):
                raise ValueError("Values in H_xt must be a positive float.")
            elif any(self.H_xt >= L_xt / 2.0):
                raise ValueError(
                    "2 * H_xt in some dimension is larger than the "
                    "corresponding grid dimension."
                )
        else:
            self.H_xt = L_xt / 20.0

        if self.spatiotemporal_grid is not None:
            if self.p < 0:
                raise ValueError("Poly degree of the spatial weights must be > 0")
            if self.p < self.derivative_order:
                self.p = self.derivative_order
        if self.K <= 0:
            raise ValueError("The number of subdomains must be > 0")

        self._set_up_weights()

    def _get_spatial_endpoints(self):
        x1 = AxesArray(np.zeros(self.grid_ndim), {"ax_coord": 0})
        x2 = AxesArray(np.zeros(self.grid_ndim), {"ax_coord": 0})
        for i in range(self.grid_ndim):
            inds = [slice(None)] * (self.grid_ndim + 1)
            for j in range(self.grid_ndim):
                inds[j] = 0
            x1[i] = self.spatiotemporal_grid[tuple(inds)][i]
            inds[i] = -1
            x2[i] = self.spatiotemporal_grid[tuple(inds)][i]
        return x1, x2

    def _set_up_weights(self):
        """
        Sets up weights needed for the weak library. Integrals over domain cells are
        approximated as dot products of weights and the input data.
        """
        dims = self.spatiotemporal_grid.shape[:-1]
        self.grid_dims = dims

        # Sample the random domain centers
        xt1, xt2 = self._get_spatial_endpoints()
        domain_centers = AxesArray(
            np.zeros((self.K, self.grid_ndim)), {"ax_sample": 0, "ax_coord": 1}
        )
        for i in range(self.grid_ndim):
            domain_centers[:, i] = np.random.uniform(
                xt1[i] + self.H_xt[i], xt2[i] - self.H_xt[i], size=self.K
            )

        # Indices for space-time points that lie in the domain cells
        self.inds_k = []
        k = 0
        while k < self.K:
            inds = []
            for i in range(self.grid_ndim):
                s = [0] * (self.grid_ndim + 1)
                s[i] = slice(None)
                s[-1] = i
                ax_vals = self.spatiotemporal_grid[tuple(s)]
                cell_left = domain_centers[k][i] - self.H_xt[i]
                cell_right = domain_centers[k][i] + self.H_xt[i]
                newinds = AxesArray(
                    ((ax_vals > cell_left) & (ax_vals < cell_right)).nonzero()[0],
                    ax_vals.axes,
                )
                # If less than two indices along any axis, resample
                if len(newinds) < 2:
                    for i in range(self.grid_ndim):
                        domain_centers[k, i] = np.random.uniform(
                            xt1[i] + self.H_xt[i], xt2[i] - self.H_xt[i], size=1
                        )
                    include = False
                    break
                else:
                    include = True
                    inds = inds + [newinds]
            if include:
                self.inds_k = self.inds_k + [inds]
                k = k + 1

        # TODO: fix meaning of axes in XT_k
        # Values of the spatiotemporal grid on the domain cells
        XT_k = [
            self.spatiotemporal_grid[np.ix_(*self.inds_k[k])] for k in range(self.K)
        ]

        # Recenter and shrink the domain cells so that grid points lie at the boundary
        # and calculate the new size
        H_xt_k = np.zeros((self.K, self.grid_ndim))
        for k in range(self.K):
            for axis in range(self.grid_ndim):
                s = [0] * (self.grid_ndim + 1)
                s[axis] = slice(None)
                s[-1] = axis
                H_xt_k[k, axis] = (XT_k[k][tuple(s)][-1] - XT_k[k][tuple(s)][0]) / 2
                domain_centers[k][axis] = (
                    XT_k[k][tuple(s)][-1] + XT_k[k][tuple(s)][0]
                ) / 2
        # Rescaled space-time values for integration weights
        xtilde_k = [(XT_k[k] - domain_centers[k]) / H_xt_k[k] for k in range(self.K)]

        # Shapes of the grid restricted to each cell
        shapes_k = np.array(
            [
                [len(self.inds_k[k][i]) for i in range(self.grid_ndim)]
                for k in range(self.K)
            ]
        )

        # Below we calculate the weights to convert integrals into dot products
        # To speed up evaluations, we proceed in several steps

        # Since the grid is a tensor product grid, we calculate weights along each axis
        # Later, we multiply the weights along each axis to produce the full weights

        # Within each domain cell, we calculate the interior weights
        # and the weights at the left and right boundaries separately,
        # since the  expression differ at the boundaries of the domains

        # Extract the space-time coordinates for each domain and the indices for
        # the left-most and right-most points for each domain.
        # We stack the values for each domain cell into a single vector to speed up
        grids = []  # the rescaled coordinates for each domain
        lefts = []  # the spatiotemporal indices at the left of each domain
        rights = []  # the spatiotemporal indices at the right of each domain
        for i in range(self.grid_ndim):
            s = [0] * (self.grid_ndim + 1)
            s[-1] = i
            s[i] = slice(None)
            # stacked coordinates for axis i over all domains
            grids = grids + [np.hstack([xtilde_k[k][tuple(s)] for k in range(self.K)])]
            # stacked indices for right-most point for axis i over all domains
            rights = rights + [np.cumsum(shapes_k[:, i]) - 1]
            # stacked indices for left-most point for axis i over all domains
            lefts = lefts + [np.concatenate([[0], np.cumsum(shapes_k[:, i])[:-1]])]

        # Weights for the time integrals along each axis
        tweights = []
        deriv = np.zeros(self.grid_ndim)
        deriv[-1] = 1
        for i in range(self.grid_ndim):
            # weights for interior points
            tweights = tweights + [self._linear_weights(grids[i], deriv[i], self.p)]
            # correct the values for the left-most points
            tweights[i][lefts[i]] = self._left_weights(
                grids[i][lefts[i]],
                grids[i][lefts[i] + 1],
                deriv[i],
                self.p,
            )
            # correct the values for the right-most points
            tweights[i][rights[i]] = self._right_weights(
                grids[i][rights[i] - 1],
                grids[i][rights[i]],
                deriv[i],
                self.p,
            )

        # Weights for pure derivative terms along each axis
        weights0 = []
        deriv = np.zeros(self.grid_ndim)
        for i in range(self.grid_ndim):
            # weights for interior points
            weights0 = weights0 + [self._linear_weights(grids[i], deriv[i], self.p)]
            # correct the values for the left-most points
            weights0[i][lefts[i]] = self._left_weights(
                grids[i][lefts[i]],
                grids[i][lefts[i] + 1],
                deriv[i],
                self.p,
            )
            # correct the values for the right-most points
            weights0[i][rights[i]] = self._right_weights(
                grids[i][rights[i] - 1],
                grids[i][rights[i]],
                deriv[i],
                self.p,
            )

        # Weights for the mixed library derivative terms along each axis
        weights1 = []
        for j in range(self.num_derivatives):
            weights2 = []
            deriv = np.concatenate([self.multiindices[j], [0]])
            for i in range(self.grid_ndim):
                # weights for interior points
                weights2 = weights2 + [self._linear_weights(grids[i], deriv[i], self.p)]
                # correct the values for the left-most points
                weights2[i][lefts[i]] = self._left_weights(
                    grids[i][lefts[i]],
                    grids[i][lefts[i] + 1],
                    deriv[i],
                    self.p,
                )
                # correct the values for the right-most points
                weights2[i][rights[i]] = self._right_weights(
                    grids[i][rights[i] - 1],
                    grids[i][rights[i]],
                    deriv[i],
                    self.p,
                )
            weights1 = weights1 + [weights2]

        # TODO: get rest of code to work with AxesArray.  Too unsure of
        # which axis labels to use at this point to continue
        tweights = [np.asarray(arr) for arr in tweights]
        weights0 = [np.asarray(arr) for arr in weights0]
        weights1 = [[np.asarray(arr) for arr in sublist] for sublist in weights1]
        # Product weights over the axes for time derivatives, shaped as inds_k
        self.fulltweights = []
        deriv = np.zeros(self.grid_ndim)
        deriv[-1] = 1
        for k in range(self.K):
            ret = np.ones(shapes_k[k])
            for i in range(self.grid_ndim):
                s = [0] * (self.grid_ndim + 1)
                s[i] = slice(None, None, None)
                s[-1] = i
                dims = np.ones(self.grid_ndim, dtype=int)
                dims[i] = shapes_k[k][i]
                ret = ret * np.reshape(
                    tweights[i][lefts[i][k] : rights[i][k] + 1], dims
                )

            self.fulltweights = self.fulltweights + [
                ret * np.prod(H_xt_k[k] ** (1.0 - deriv))
            ]

        # Product weights over the axes for pure derivative terms, shaped as inds_k
        self.fullweights0 = []
        for k in range(self.K):
            ret = np.ones(shapes_k[k])
            for i in range(self.grid_ndim):
                s = [0] * (self.grid_ndim + 1)
                s[i] = slice(None, None, None)
                s[-1] = i
                dims = np.ones(self.grid_ndim, dtype=int)
                dims[i] = shapes_k[k][i]
                ret = ret * np.reshape(
                    weights0[i][lefts[i][k] : rights[i][k] + 1], dims
                )

            self.fullweights0 = self.fullweights0 + [ret * np.prod(H_xt_k[k])]

        # Product weights over the axes for mixed derivative terms, shaped as inds_k
        self.fullweights1 = []
        for k in range(self.K):
            weights2 = []
            for j in range(self.num_derivatives):
                if not self.implicit_terms:
                    deriv = np.concatenate([self.multiindices[j], [0]])
                else:
                    deriv = self.multiindices[j]

                ret = np.ones(shapes_k[k])
                for i in range(self.grid_ndim):
                    s = [0] * (self.grid_ndim + 1)
                    s[i] = slice(None, None, None)
                    s[-1] = i
                    dims = np.ones(self.grid_ndim, dtype=int)
                    dims[i] = shapes_k[k][i]
                    ret = ret * np.reshape(
                        weights1[j][i][lefts[i][k] : rights[i][k] + 1],
                        dims,
                    )

                weights2 = weights2 + [ret * np.prod(H_xt_k[k] ** (1.0 - deriv))]
            self.fullweights1 = self.fullweights1 + [weights2]

    def _phi(self, x, d, p):
        """
        One-dimensional polynomial test function (1-x**2)**p,
        differentiated d times, calculated term-wise in the binomial
        expansion.
        """
        ks = np.arange(self.p + 1)
        ks = ks[np.where(2 * (self.p - ks) - d >= 0)][:, np.newaxis]
        return np.sum(
            binom(self.p, ks)
            * (-1) ** ks
            * x[np.newaxis, :] ** (2 * (self.p - ks) - d)
            * perm(2 * (self.p - ks), d),
            axis=0,
        )

    def _phi_int(self, x, d, p):
        """
        Indefinite integral of one-dimensional polynomial test
        function (1-x**2)**p, differentiated d times, calculated
        term-wise in the binomial expansion.
        """
        ks = np.arange(self.p + 1)
        ks = ks[np.where(2 * (self.p - ks) - d >= 0)][:, np.newaxis]
        return np.sum(
            binom(self.p, ks)
            * (-1) ** ks
            * x[np.newaxis, :] ** (2 * (self.p - ks) - d + 1)
            * perm(2 * (self.p - ks), d)
            / (2 * (self.p - ks) - d + 1),
            axis=0,
        )

    def _xphi_int(self, x, d, p):
        """
        Indefinite integral of one-dimensional polynomial test function
        x*(1-x**2)**p, differentiated d times, calculated term-wise in the
        binomial expansion.
        """
        ks = np.arange(self.p + 1)
        ks = ks[np.where(2 * (self.p - ks) - d >= 0)][:, np.newaxis]
        return np.sum(
            binom(self.p, ks)
            * (-1) ** ks
            * x[np.newaxis, :] ** (2 * (self.p - ks) - d + 2)
            * perm(2 * (self.p - ks), d)
            / (2 * (self.p - ks) - d + 2),
            axis=0,
        )

    def _linear_weights(self, x, d, p):
        """
        One-dimensioal weights for integration against the dth derivative
        of the polynomial test function (1-x**2)**p. This is derived
        assuming the function to integrate is linear between grid points:
        f(x)=f_i+(x-x_i)/(x_{i+1}-x_i)*(f_{i+1}-f_i)
        so that f(x)*dphi(x) is a piecewise polynomial.
        The piecewise components are computed analytically, and the integral is
        expressed as a dot product of weights against the f_i.
        """
        ws = self._phi_int(x, d, p)
        zs = self._xphi_int(x, d, p)
        return np.concatenate(
            [
                [
                    x[1] / (x[1] - x[0]) * (ws[1] - ws[0])
                    - 1 / (x[1] - x[0]) * (zs[1] - zs[0])
                ],
                x[2:] / (x[2:] - x[1:-1]) * (ws[2:] - ws[1:-1])
                - x[:-2] / (x[1:-1] - x[:-2]) * (ws[1:-1] - ws[:-2])
                + 1 / (x[1:-1] - x[:-2]) * (zs[1:-1] - zs[:-2])
                - 1 / (x[2:] - x[1:-1]) * (zs[2:] - zs[1:-1]),
                [
                    -x[-2] / (x[-1] - x[-2]) * (ws[-1] - ws[-2])
                    + 1 / (x[-1] - x[-2]) * (zs[-1] - zs[-2])
                ],
            ]
        )

    def _left_weights(self, x1, x2, d, p):
        """
        One-dimensioal weight for left-most point in integration against the dth
        derivative of the polynomial test function (1-x**2)**p. This is derived
        assuming the function to integrate is linear between grid points:
        f(x)=f_i+(x-x_i)/(x_{i+1}-x_i)*(f_{i+1}-f_i)
        so that f(x)*dphi(x) is a piecewise polynomial.
        The piecewise components are computed analytically, and the integral is
        expressed as a dot product of weights against the f_i.
        """
        w1 = self._phi_int(x1, d, p)
        w2 = self._phi_int(x2, d, p)
        z1 = self._xphi_int(x1, d, p)
        z2 = self._xphi_int(x2, d, p)
        return x2 / (x2 - x1) * (w2 - w1) - 1 / (x2 - x1) * (z2 - z1)

    def _right_weights(self, x1, x2, d, p):
        """
        One-dimensioal weight for right-most point in integration against the dth
        derivative of the polynomial test function (1-x**2)**p. This is derived
        assuming the function to integrate is linear between grid points:
        f(x)=f_i+(x-x_i)/(x_{i+1}-x_i)*(f_{i+1}-f_i)
        so that f(x)*dphi(x) is a piecewise polynomial.
        The piecewise components are computed analytically, and the integral is
        expressed as a dot product of weights against the f_i.
        """
        w1 = self._phi_int(x1, d, p)
        w2 = self._phi_int(x2, d, p)
        z1 = self._xphi_int(x1, d, p)
        z2 = self._xphi_int(x2, d, p)
        return -x1 / (x2 - x1) * (w2 - w1) + 1 / (x2 - x1) * (z2 - z1)

    def convert_u_dot_integral(self, u):
        """
        Takes a full set of spatiotemporal fields u(x, t) and finds the weak
        form of u_dot.
        """
        K = self.K
        gdim = self.grid_ndim
        u_dot_integral = np.zeros((K, u.shape[-1]))
        deriv_orders = np.zeros(gdim)
        deriv_orders[-1] = 1

        # Extract the input features on indices in each domain cell
        dims = np.array(self.spatiotemporal_grid.shape)
        dims[-1] = u.shape[-1]

        for k in range(self.K):  # loop over domain cells
            # calculate the integral feature by taking the dot product
            # of the weights and functions over each axis
            u_dot_integral[k] = np.tensordot(
                self.fulltweights[k],
                -u[np.ix_(*self.inds_k[k])],
                axes=(
                    tuple(np.arange(self.grid_ndim)),
                    tuple(np.arange(self.grid_ndim)),
                ),
            )

        return u_dot_integral

    def get_feature_names(self, input_features=None):
        """Return feature names for output features.

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        check_is_fitted(self)
        n_features = self.n_features_in_
        if input_features is None:
            input_features = ["x%d" % i for i in range(n_features)]
        feature_names = []
        lib_names = []

        # Include constant term
        if self.include_bias:
            feature_names.append("1")
        # Include any non-derivative terms
        lib_names = self.function_library.get_feature_names(input_features)
        feature_names = feature_names + lib_names

        if self.grid_ndim != 0:

            def derivative_string(multiindex):
                ret = ""
                for axis in range(self.ind_range):
                    if (axis == self.ind_range - 1) and (
                        self.ind_range == self.grid_ndim
                    ):
                        str_deriv = "t"
                    else:
                        str_deriv = str(axis + 1)
                    for i in range(multiindex[axis]):
                        ret = ret + str_deriv
                return ret

            # Include integral terms
            for k in range(self.num_derivatives):
                for j in range(n_features):
                    feature_names.append(
                        input_features[j]
                        + "_"
                        + derivative_string(self.multiindices[k])
                    )
            # Include mixed non-derivative + integral terms
            if self.include_interaction:
                for k in range(self.num_derivatives):
                    for jj in range(n_features):
                        for lib_name in lib_names:
                            feature_names.append(
                                lib_name
                                + input_features[jj]
                                + "_"
                                + derivative_string(self.multiindices[k])
                            )
        return feature_names

    @x_sequence_or_item
    def fit(self, x_full, y=None):
        """Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            Measurement data.

        Returns
        -------
        self : instance
        """
        x0 = x_full[0]
        n_features = x0.shape[x0.ax_coord]
        self.n_features_in_ = n_features
        n_output_features = 0

        # Count the number of non-derivative terms
        self.function_library.fit(x0.take(0, x0.ax_time))
        n_output_features = self.function_library.n_output_features_

        if self.grid_ndim != 0:
            # Add the mixed derivative library_terms
            if self.include_interaction:
                n_output_features += (
                    n_output_features * n_features * self.num_derivatives
                )
            # Add the pure derivative library terms
            n_output_features += n_features * self.num_derivatives

        # If there is a constant term, add 1 to n_output_features
        if self.include_bias:
            n_output_features += 1

        self.n_output_features_ = n_output_features

        # required to generate the function names
        self.get_feature_names()

        return self

    @x_sequence_or_item
    def transform(self, x_full):
        """Transform data to custom features

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray, shape (n_samples, n_output_features)
            The matrix of features, where n_output_features is the number of
            features generated from applying the custom functions
            to the inputs.
        """
        check_is_fitted(self)

        xp_full = []
        for x in x_full:
            n_features = x.shape[x.ax_coord]
            xp = np.empty((self.K, self.n_output_features_), dtype=x.dtype)

            # Extract the input features on indices in each domain cell
            self.x_k = [x[np.ix_(*self.inds_k[k])] for k in range(self.K)]

            # library function terms

            # Evaluate the functions on the indices of domain cells
            funcs = self.function_library.fit_transform(x)
            n_library_terms = funcs.shape[-1]
            library_functions = np.empty((self.K, n_library_terms), dtype=x.dtype)

            # library function terms
            for k in range(self.K):  # loop over domain cells
                # calculate the integral feature by taking the dot product
                # of the weights and functions over each axis
                library_functions[k] = np.tensordot(
                    self.fullweights0[k],
                    funcs[np.ix_(*self.inds_k[k])],
                    axes=(
                        tuple(np.arange(self.grid_ndim)),
                        tuple(np.arange(self.grid_ndim)),
                    ),
                )

            if self.derivative_order != 0:
                # pure integral terms
                library_integrals = np.empty(
                    (self.K, n_features * self.num_derivatives), dtype=x.dtype
                )

                for k in range(self.K):  # loop over domain cells
                    library_idx = 0
                    for j in range(self.num_derivatives):  # loop over derivatives
                        # Calculate the integral feature by taking the dot product
                        # of the weights and data x_k over each axis.
                        # Integration by parts gives power of (-1).
                        library_integrals[k, library_idx : library_idx + n_features] = (
                            -1
                        ) ** (np.sum(self.multiindices[j])) * np.tensordot(
                            self.fullweights1[k][j],
                            self.x_k[k],
                            axes=(
                                tuple(np.arange(self.grid_ndim)),
                                tuple(np.arange(self.grid_ndim)),
                            ),
                        )
                        library_idx += n_features

                # Mixed derivative/non-derivative terms
                if self.include_interaction:
                    library_mixed_integrals = np.empty(
                        (
                            self.K,
                            n_library_terms * n_features * self.num_derivatives,
                        ),
                        dtype=x.dtype,
                    )

                    # Below we integrate the product of function and feature
                    # derivatives against the derivatives of phi to calculate the weak
                    # features. We cannot remove all derivatives of data in this case,
                    # but we can reduce the derivative order by half.

                    # Calculate the necessary function and feature derivatives
                    funcs_derivs = np.zeros(
                        np.concatenate([[self.num_derivatives + 1], funcs.shape])
                    )
                    x_derivs = np.zeros(
                        np.concatenate([[self.num_derivatives + 1], x.shape])
                    )
                    funcs_derivs[0] = funcs
                    x_derivs[0] = x
                    self.dx_k_j = []
                    self.dfx_k_j = []

                    for j in range(self.num_derivatives):
                        funcs_derivs[j + 1] = funcs
                        x_derivs[j + 1] = x
                        for axis in range(self.ind_range):
                            s = [0] * (self.grid_ndim + 1)
                            s[axis] = slice(None, None, None)
                            s[-1] = axis
                            # Need derivatives of order less than half derivative_order
                            if self.multiindices[j][axis] > 0 and self.multiindices[j][
                                axis
                            ] <= (self.derivative_order // 2):
                                funcs_derivs[j + 1] = self.differentiation_method(
                                    d=self.multiindices[j][axis],
                                    axis=axis,
                                    **self.diff_kwargs,
                                )._differentiate(
                                    funcs_derivs[j + 1],
                                    self.spatiotemporal_grid[tuple(s)],
                                )
                            if self.multiindices[j][axis] > 0 and self.multiindices[j][
                                axis
                            ] <= (self.derivative_order - (self.derivative_order // 2)):
                                x_derivs[j + 1] = self.differentiation_method(
                                    d=self.multiindices[j][axis],
                                    axis=axis,
                                    **self.diff_kwargs,
                                )._differentiate(
                                    x_derivs[j + 1], self.spatiotemporal_grid[tuple(s)]
                                )

                    # Extract the function and feature derivatives on the domains
                    self.dx_k_j = [
                        [
                            x_derivs[j][np.ix_(*self.inds_k[k])]
                            for j in range(self.num_derivatives + 1)
                        ]
                        for k in range(self.K)
                    ]
                    self.dfx_k_j = [
                        [
                            funcs_derivs[j][np.ix_(*self.inds_k[k])]
                            for j in range(self.num_derivatives + 1)
                        ]
                        for k in range(self.K)
                    ]

                    # Calculate the mixed integrals
                    library_idx = 0
                    for j in range(self.num_derivatives):
                        integral = np.zeros((self.K, n_library_terms, n_features))
                        # Derivative orders after integration by parts
                        derivs_mixed = self.multiindices[j] // 2
                        derivs_pure = self.multiindices[j] - derivs_mixed
                        # Derivative orders for mixed derivatives product rule
                        derivs = np.concatenate(
                            [
                                [np.zeros(self.ind_range, dtype=int)],
                                self.multiindices,
                            ],
                            axis=0,
                        )
                        # Sum the terms in product rule
                        for deriv in derivs[
                            np.where(np.all(derivs <= derivs_mixed, axis=1))[0]
                        ]:
                            # indices for product rule terms
                            j0 = np.where(np.all(derivs == deriv, axis=1))[0][0]
                            j1 = np.where(
                                np.all(derivs == derivs_mixed - deriv, axis=1)
                            )[0][0]
                            j2 = np.where(np.all(derivs == derivs_pure, axis=1))[0][0]

                            for k in range(self.K):
                                # Weights are either in fullweights0 or fullweights1
                                if j0 == 0:
                                    weights = self.fullweights0[k]
                                else:
                                    weights = self.fullweights1[k][j0 - 1]

                                # Calculate the integral by taking the dot product
                                # of the weights and data x_k over each axis.
                                # Integration by parts gives power of (-1).
                                # Binomial factor comes by product rule.
                                integral[k] = integral[k] + (-1) ** (
                                    np.sum(derivs_mixed)
                                ) * np.tensordot(
                                    weights,
                                    self.dfx_k_j[k][j1][..., np.newaxis]
                                    * self.dx_k_j[k][j2][..., np.newaxis, :],
                                    axes=(
                                        tuple(np.arange(self.grid_ndim)),
                                        tuple(np.arange(self.grid_ndim)),
                                    ),
                                ) * np.prod(
                                    binom(derivs_mixed, deriv)
                                )
                        # collect the results
                        for n in range(n_features):
                            for m in range(n_library_terms):
                                library_mixed_integrals[:, library_idx] = integral[
                                    :, m, n
                                ]
                                library_idx += 1

            library_idx = 0
            # Constant term
            if self.include_bias:
                constants_final = np.zeros(self.K)
                for k in range(self.K):
                    constants_final[k] = np.sum(self.fullweights0[k])
                xp[:, library_idx] = constants_final
                library_idx += 1

            # library function terms
            xp[:, library_idx : library_idx + n_library_terms] = library_functions
            library_idx += n_library_terms

            if self.derivative_order != 0:
                # pure integral terms
                xp[
                    :, library_idx : library_idx + self.num_derivatives * n_features
                ] = library_integrals
                library_idx += self.num_derivatives * n_features

                # mixed function integral terms
                if self.include_interaction:
                    xp[
                        :,
                        library_idx : library_idx
                        + n_library_terms * self.num_derivatives * n_features,
                    ] = library_mixed_integrals
                    library_idx += n_library_terms * self.num_derivatives * n_features

            xp_full = xp_full + [AxesArray(xp, {"ax_sample": 0, "ax_coord": 1})]
        return xp_full

    def calc_trajectory(self, diff_method, x, t):
        x_dot = self.convert_u_dot_integral(x)
        return x, AxesArray(x_dot, {"ax_sample": 0, "ax_coord": 1})



import warnings

import numpy as np
from sklearn.linear_model import ridge_regression

from .base import BaseOptimizer


class FROLS(BaseOptimizer):
    """Forward Regression Orthogonal Least-Squares (FROLS) optimizer.

    Attempts to minimize the objective function
    :math:`\\|y - Xw\\|^2_2 + \\alpha \\|w\\|^2_2`
    by iteractively selecting the most correlated
    function in the library. This is a greedy algorithm.

    See the following reference for more details:

        Billings, Stephen A. Nonlinear system identification:
        NARMAX methods in the time, frequency, and spatio-temporal domains.
        John Wiley & Sons, 2013.

    Parameters
    ----------
    normalize_columns : boolean, optional (default False)
        Normalize the columns of x (the SINDy library terms) before regression
        by dividing by the L2-norm. Note that the 'normalize' option in sklearn
        is deprecated in sklearn versions >= 1.0 and will be removed.

    copy_X : boolean, optional (default True)
        If True, X will be copied; else, it may be overwritten.

    kappa : float, optional (default None)
        If passed, compute the MSE errors with an extra L0 term with
        strength equal to kappa times the condition number of Theta.

    max_iter : int, optional (default 10)
        Maximum iterations of the optimization algorithm. This determines
        the number of nonzero terms chosen by the FROLS algorithm.

    alpha : float, optional (default 0.05)
        Optional L2 (ridge) regularization on the weight vector.

    ridge_kw : dict, optional (default None)
        Optional keyword arguments to pass to the ridge regression.

    verbose : bool, optional (default False)
        If True, prints out the different error terms every
        iteration.

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s).

    history_ : list
        History of ``coef_``. ``history_[k]`` contains the values of
        ``coef_`` at iteration k of FROLS.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.integrate import odeint
    >>> from pysindy import SINDy
    >>> from pysindy.optimizers import FROLS
    >>> lorenz = lambda z,t : [10 * (z[1] - z[0]),
    >>>                        z[0] * (28 - z[2]) - z[1],
    >>>                        z[0] * z[1] - 8 / 3 * z[2]]
    >>> t = np.arange(0, 2, .002)
    >>> x = odeint(lorenz, [-8, 8, 27], t)
    >>> opt = FROLS(threshold=.1, alpha=.5)
    >>> model = SINDy(optimizer=opt)
    >>> model.fit(x, t=t[1] - t[0])
    >>> model.print()
    x0' = -9.999 1 + 9.999 x0
    x1' = 27.984 1 + -0.996 x0 + -1.000 1 x1
    x2' = -2.666 x1 + 1.000 1 x0
    """

    def __init__(
        self,
        normalize_columns=False,
        copy_X=True,
        kappa=None,
        max_iter=10,
        alpha=0.05,
        ridge_kw=None,
        verbose=False,
        unbias=True,
    ):
        super().__init__(
            copy_X=copy_X,
            max_iter=max_iter,
            normalize_columns=normalize_columns,
            unbias=unbias,
        )
        self.alpha = alpha
        self.ridge_kw = ridge_kw
        self.kappa = kappa
        if self.max_iter <= 0:
            raise ValueError("Max iteration must be > 0")
        self.verbose = verbose

    def _normed_cov(self, a, b):
        with warnings.catch_warnings():
            warnings.filterwarnings("error", category=RuntimeWarning)
            try:
                return np.vdot(a, b) / np.vdot(a, a)
            except RuntimeWarning:
                raise ValueError(
                    "Trying to orthogonalize linearly dependent columns created NaNs"
                )

    def _select_function(self, x, y, sigma, skip=[]):
        n_features = x.shape[1]
        g = np.zeros(n_features)  # Coefficients to orthogonalized functions
        ERR = np.zeros(n_features)  # Error Reduction Ratio at this step
        for m in range(n_features):
            if m not in skip:
                g[m] = self._normed_cov(x[:, m], y)
                ERR[m] = (
                    abs(g[m]) ** 2 * np.real(np.vdot(x[:, m], x[:, m])) / sigma
                )  # Error reduction

        best_idx = np.argmax(
            ERR
        )  # Select best function by maximum Error Reduction Ratio

        # Return index of best function, along with ERR and coefficient
        return best_idx, ERR[best_idx], g[best_idx]

    def _orthogonalize(self, vec, Q):
        """
        Orthogonalize vec with respect to columns of Q
        """
        Qs = vec.copy()
        s = Q.shape[1]
        for r in range(s):
            Qs -= self._normed_cov(Q[:, r], Qs) * Q[:, r]
        return Qs

    def _reduce(self, x, y):
        """
        Performs at most n_feature iterations of the
        greedy Forward Regression Orthogonal Least Squares (FROLS) algorithm
        """
        n_samples, n_features = x.shape
        n_targets = y.shape[1]
        cond_num = np.linalg.cond(x)
        if self.kappa is not None:
            l0_penalty = self.kappa * cond_num
        else:
            l0_penalty = 0.0

        # Print initial values for each term in the optimization
        if self.verbose:
            row = [
                "Iteration",
                "Index",
                "|y - Xw|^2",
                "a * |w|_2",
                "|w|_0",
                "b * |w|_0",
                "Total: |y-Xw|^2+a*|w|_2+b*|w|_0",
            ]
            print(
                "{: >10} ... {: >5} ... {: >10} ... {: >10} ... {: >5}"
                " ... {: >10} ... {: >10}".format(*row)
            )
            print(
                " Note that these error metrics are related but not the same"
                " as the loss function being minimized by FROLS!"
            )

        # History of selected functions as a list of [output x coefficients]
        # with a length of number of iterations
        self.history_ = []
        # Error Reduction Ratio: [iteration x output]
        self.ERR_global = np.zeros((n_features, n_targets))
        # Initialize arrays
        g_global_full = np.zeros(
            (n_targets, n_features), dtype=x.dtype
        )  # Coefficients for selected (orthogonal) functions
        L_full = np.zeros(
            (n_targets, n_features), dtype=int
        )  # Order of selection, i.e. L[0] is the first, L[1] second...
        A_full = np.zeros(
            (n_targets, n_features, n_features), dtype=x.dtype
        )  # Used for inversion to original function set (A @ coef = g)
        # Orthogonal function libraries
        Q_full = np.zeros(
            (n_targets, *x.shape), dtype=x.dtype
        )  # Global library (built over time)
        Qs_full = np.zeros(
            (n_targets, *x.shape), dtype=x.dtype
        )  # Same, but for each step
        for i in range(n_features):
            coef_i = np.zeros((n_targets, n_features), dtype=x.dtype)

            for k in range(n_targets):
                g_global = g_global_full[k]
                L = L_full[k]
                A = A_full[k]
                Q = Q_full[k]
                Qs = Qs_full[k]
                sigma = np.real(np.vdot(y[:, k], y[:, k]))  # Variance of the signal

                for m in range(n_features):
                    if m not in L[:i]:
                        # Orthogonalize with respect to already selected functions
                        Qs[:, m] = self._orthogonalize(x[:, m], Q[:, :i])

                L[i], self.ERR_global[i, k], g_global[i] = self._select_function(
                    Qs, y[:, k], sigma, L[:i]
                )

                # Store transformation from original functions to orthogonal library
                for j in range(i):
                    A[j, i] = self._normed_cov(Q[:, j], x[:, L[i]])
                A[i, i] = 1.0
                Q[:, i] = Qs[:, L[i]].copy()
                Qs *= 0.0

                # Invert orthogonal coefficient vector
                # to get coefficients for original functions
                # at this step of the iteration
                if i != 0:
                    kw = self.ridge_kw or {}
                    alpha = ridge_regression(A[:i, :i], g_global[:i], self.alpha, **kw)
                else:
                    alpha = []

                # Coefficients for the library functions
                coef_k = np.zeros_like(g_global)
                coef_k[L[:i]] = alpha
                coef_k[abs(coef_k) < 1e-10] = 0
                coef_i[k, :] = coef_k

                if self.verbose:
                    coef = coef_i[k, :]
                    R2 = np.sum((y[:, k] - np.dot(x, coef).T) ** 2)
                    L2 = self.alpha * np.sum(coef**2)
                    L0 = np.count_nonzero(coef)
                    row = [i, k, R2, L2, L0, l0_penalty * L0, R2 + L2 + l0_penalty * L0]
                    print(
                        "{0:10d} ... {1:5d} ... {2:10.4e} ... {3:10.4e}"
                        " ... {4:5d} ... {5:10.4e} ... {6:10.4e}".format(*row)
                    )
            self.history_.append(coef_i)
            if i >= self.max_iter:
                break

        # Function selection: L2 error for output k at iteration i is given by
        # sum(ERR_global[k, :i]), and the number of nonzero coefficients is (i+1)
        l2_err = np.cumsum(self.ERR_global, axis=1)
        l0_norm = np.arange(1, n_features + 1)[:, None]
        self.loss_ = l2_err + l0_penalty * l0_norm  # Save for debugging

        # For each output, choose the minimum L0-penalized loss function
        self.coef_ = np.zeros((n_targets, n_features), dtype=x.dtype)
        loss_min = np.argmin(
            self.loss_[: self.max_iter + 1, :], axis=0
        )  # Minimum loss for this output function
        for k in range(n_targets):
            self.coef_[k, :] = self.history_[loss_min[k]][k, :]



"""
Base class for SINDy optimizers.
"""
import abc
import warnings
from typing import Callable
from typing import NewType
from typing import Optional
from typing import Tuple

import numpy as np
from scipy import sparse
from sklearn.base import BaseEstimator
from sklearn.linear_model import LinearRegression
from sklearn.utils import check_array
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import check_X_y

from .._typing import Float2D
from .._typing import FloatDType
from ..utils import AxesArray
from ..utils import drop_nan_samples

NFeat = NewType("NFeat", int)
NTarget = NewType("NTarget", int)


def _rescale_data(X, y, sample_weight):
    """Rescale data so as to support sample_weight"""
    n_samples = X.shape[0]
    sample_weight = np.asarray(sample_weight)
    if sample_weight.ndim == 0:
        sample_weight = np.full(n_samples, sample_weight, dtype=sample_weight.dtype)
    sample_weight_sr = np.sqrt(sample_weight)
    sw_matrix = sparse.dia_matrix((sample_weight_sr, 0), shape=(n_samples, n_samples))
    X = safe_sparse_dot(sw_matrix, X)
    y = safe_sparse_dot(sw_matrix, y)
    return X, y, sample_weight_sr


def _preprocess_data(
    X,
    y,
    *,
    fit_intercept: bool = False,
    copy=True,
    copy_y=True,
    sample_weight=None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, None, Optional[np.ndarray]]:
    """A clone/variant of sklearn's _preprocess_data, but that's private

    It makes assumptions that are always true for pysindy but may not be for
    generic sklearn estimators.

    In future, will harness power of sklearn's array API compatibility layer
    to determine device, array namespace (e.g. jax).
    """
    X = check_array(X, copy=copy, accept_sparse=["csr", "csc"])
    y = check_array(y, dtype=X.dtype, copy=copy_y, ensure_2d=False)

    n_samples, n_features = X.shape

    dtype_ = X.dtype
    X_offset = np.zeros(n_features, dtype=X.dtype)
    if y.ndim == 1:
        y_offset = np.asarray(0.0, dtype=dtype_)
    else:
        y_offset = np.zeros(y.shape[1], dtype=dtype_)
    if sample_weight is not None:
        X, y, sample_weight_sr = _rescale_data(X, y, sample_weight)
    else:
        sample_weight_sr = None

    return X, y, X_offset, y_offset, None, sample_weight_sr


class _BaseOptimizer(BaseEstimator, abc.ABC):
    coef_: np.ndarray[tuple[NTarget, NFeat], FloatDType]
    intercept_: np.ndarray[tuple[NTarget], FloatDType]

    @property
    def complexity(self):
        check_is_fitted(self)
        return np.count_nonzero(self.coef_) + np.count_nonzero(self.intercept_)


class BaseOptimizer(LinearRegression, _BaseOptimizer):
    """
    Base class for SINDy optimizers. Subclasses must implement
    a _reduce method for carrying out the bulk of the work of
    fitting a model.

    Parameters
    ----------
    normalize_columns : boolean, optional (default False)
        Normalize the columns of x (the SINDy library terms) before regression
        by dividing by the L2-norm.

    copy_X : boolean, optional (default True)
        If True, X will be copied; else, it may be overwritten.

    initial_guess : np.ndarray, shape (n_features,) or (n_targets, n_features),
            optional (default None)
        Initial guess for coefficients ``coef_``.
        If None, the initial guess is obtained via a least-squares fit.

    unbias:  Whether to perform an extra step of unregularized linear
        regression to unbias the coefficients for the identified
        support.  If an optimizer (``self.optimizer``) applies any type
        of regularization, that regularization may bias coefficients,
        improving the conditioning of the problem but harming the
        quality of the fit. Setting ``unbias==True`` enables an extra
        step wherein unregularized linear regression is applied, but
        only for the coefficients in the support identified by the
        optimizer. This helps to remove the bias introduced by
        regularization.

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s).

    ind_ : array, shape (n_features,) or (n_targets, n_features)
        Array of bools indicating which coefficients of the
        weight vector have not been masked out.

    history_ : list
        History of ``coef_`` over iterations of the optimization algorithm.

    Theta_ : np.ndarray, shape (n_samples, n_features)
        The Theta matrix to be used in the optimization. We save it as
        an attribute because access to the full library of terms is
        sometimes needed for various applications.

    """

    max_iter: int
    normalize_columns: bool
    initial_guess: Optional[np.ndarray[tuple[NTarget, NFeat], FloatDType]]
    copy_X: bool
    unbias: bool

    def __init__(
        self,
        max_iter=20,
        normalize_columns=False,
        initial_guess=None,
        copy_X=True,
        unbias: bool = True,
    ):
        super().__init__(fit_intercept=False, copy_X=copy_X)
        self.max_iter = max_iter
        self.iters = 0
        self.initial_guess = initial_guess
        self.normalize_columns = normalize_columns
        self.unbias = unbias
        self.__post_init_guard()

    # See name mangling rules for double underscore rationale
    def __post_init_guard(self):
        """Conduct initialization post-init, as required by scikitlearn API."""
        if np.ndim(self.initial_guess) == 1:
            self.initial_guess = self.initial_guess.reshape(1, -1)
        if self.max_iter <= 0:
            raise ValueError("max_iter must be positive")

    def set_params(self, **kwargs):
        super().set_params(**kwargs)
        self.__post_init_guard

    # Force subclasses to implement this
    @abc.abstractmethod
    def _reduce(self):
        """
        Carry out the bulk of the work of the fit function.

        Subclass implementations MUST update self.coef_ as shape
            (n_targets, n_inputs).
        """
        raise NotImplementedError

    def fit(self, x_, y, sample_weight=None, **reduce_kws):
        """
        Fit the model.

        Parameters
        ----------
        x_ : array-like, shape (n_samples, n_features)
            Training data

        y : array-like, shape (n_samples,) or (n_samples, n_targets)
            Target values

        sample_weight : float or numpy array of shape (n_samples,), optional
            Individual weights for each sample

        reduce_kws : dict
            Optional keyword arguments to pass to the _reduce method
            (implemented by subclasses)

        Returns
        -------
        self : returns an instance of self
        """
        x_ = AxesArray(np.asarray(x_), {"ax_sample": 0, "ax_coord": 1})
        y_axes = {"ax_sample": 0} if y.ndim == 1 else {"ax_sample": 0, "ax_coord": 1}
        y = AxesArray(np.asarray(y), y_axes)
        x_, y = drop_nan_samples(x_, y)
        x_, y = check_X_y(x_, y, accept_sparse=[], y_numeric=True, multi_output=True)

        x, y, X_offset, y_offset, _, sample_weight_sqrt = _preprocess_data(
            x_,
            y,
            fit_intercept=False,
            copy=self.copy_X,
            sample_weight=sample_weight,
        )

        self.iters = 0

        if y.ndim == 1:
            y = y.reshape(-1, 1)
        coef_shape = (y.shape[1], x.shape[1])
        self.ind_ = np.ones(coef_shape, dtype=bool)

        self.Theta_ = x

        x_normed = np.copy(x)
        if self.normalize_columns:
            feat_norms, x_normed = _normalize_features(x_normed)

        if self.initial_guess is None:
            self.coef_ = np.linalg.lstsq(x_normed, y, rcond=None)[0].T
        else:
            if not self.initial_guess.shape == coef_shape:
                raise ValueError(
                    "initial_guess shape is incompatible with training data. "
                    f"Expected: {coef_shape}. Received: {self.initial_guess.shape}."
                )
            self.coef_ = self.initial_guess

        self.history_ = [self.coef_]

        x_normed = np.asarray(x_normed)

        self._reduce(x_normed, y, **reduce_kws)
        self.ind_ = np.abs(self.coef_) > 1e-14

        if self.unbias:
            self._unbias(x_normed, y)

        # Rescale coefficients to original units
        if self.normalize_columns:
            self.coef_ = self.coef_ / feat_norms
            if hasattr(self, "coef_full_"):
                self.coef_full_ = self.coef_full_ / feat_norms
            for i in range(np.shape(self.history_)[0]):
                self.history_[i] = self.history_[i] / feat_norms

        # Intercept is the responsibility of feature libraries
        self.intercept_ = 0.0
        return self

    def _unbias(self, x: np.ndarray, y: np.ndarray):
        coef = np.zeros((y.shape[1], x.shape[1]))
        for i in range(self.ind_.shape[0]):
            if np.any(self.ind_[i]):
                coef[i, self.ind_[i]] = (
                    LinearRegression(fit_intercept=False)
                    .fit(x[:, self.ind_[i]], y[:, i])
                    .coef_
                )
        self.coef_ = coef


class EnsembleOptimizer(BaseOptimizer):
    """Wrapper class for ensembling methods.

    Parameters
    ----------
    opt: BaseOptimizer
        The underlying optimizer to run on each ensemble

    bagging : boolean, optional (default False)
        This parameter is used to allow for "ensembling", i.e. the
        generation of many SINDy models (n_models) by choosing a random
        temporal subset of the input data (n_subset) for each sparse
        regression. This often improves robustness because averages
        (bagging) or medians (bragging) of all the models are usually
        quite high-performing. The user can also generate "distributions"
        of many models, and calculate how often certain library terms
        are included in a model.

    library_ensemble : boolean, optional (default False)
        This parameter is used to allow for "library ensembling",
        i.e. the generation of many SINDy models (n_models) by choosing
        a random subset of the candidate library terms to truncate. So,
        n_models are generated by solving n_models sparse regression
        problems on these "reduced" libraries. Once again, this often
        improves robustness because averages (bagging) or medians
        (bragging) of all the models are usually quite high-performing.
        The user can also generate "distributions" of many models, and
        calculate how often certain library terms are included in a model.

    n_models : int, optional (default 20)
        Number of models to generate via ensemble

    n_subset : int, optional (default len(time base))
        Number of time points to use for ensemble.
        When bagging with replacement (bootstrap), a value equal to the
        original number of samples is standard.
        See: B. Efron (1979), "Bootstrap Methods: Another Look at the
        Jackknife", The Annals of Statistics.

    n_candidates_to_drop : int, optional (default 1)
        Number of candidate terms in the feature library to drop during
        library ensembling.

    replace : boolean, optional (default True)
        If ensemble true, whether or not to time sample with replacement.

    ensemble_aggregator : callable, optional (default numpy.median)
        Method to aggregate model coefficients across different samples.
        This method argument is only used if ``ensemble`` or ``library_ensemble``
        is True.
        The method should take in a list of 2D arrays and return a 2D
        array of the same shape as the arrays in the list.
        Example: :code:`lambda x: np.median(x, axis=0)`

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Regularized weight vector(s). This is the v in the objective
        function.

    coef_full_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s) that are not subjected to the regularization.
        This is the w in the objective function.
    """

    def __init__(
        self,
        opt: BaseOptimizer,
        bagging: bool = False,
        library_ensemble: bool = False,
        n_models: int = 20,
        n_subset: int = None,
        n_candidates_to_drop: int = 1,
        replace: bool = True,
        ensemble_aggregator: Callable = None,
    ):
        if not hasattr(opt, "initial_guess"):
            opt.initial_guess = None

        super().__init__(
            max_iter=opt.max_iter,
            initial_guess=opt.initial_guess,
            copy_X=opt.copy_X,
        )
        if not bagging and not library_ensemble:
            raise ValueError(
                "If not ensembling data or library terms, use another optimizer"
            )
        if bagging and n_subset is not None and n_subset < 1:
            raise ValueError("n_subset must be a positive integer or None if bagging")
        if library_ensemble and (
            n_candidates_to_drop is None or n_candidates_to_drop < 1
        ):
            raise ValueError(
                "n_candidates_to_drop must be a positive integer if ensembling library"
            )
        if n_models < 1:
            raise ValueError(
                "n_candidates_to_drop must be a positive integer if ensembling library"
            )
        self.opt = opt
        self.n_models = n_models
        self.n_subset = n_subset
        self.bagging = bagging
        self.library_ensemble = library_ensemble
        self.ensemble_aggregator = ensemble_aggregator
        self.replace = replace
        self.n_candidates_to_drop = n_candidates_to_drop
        self.coef_list = []
        self.unbias = False

    def _reduce(self, x: AxesArray, y: np.ndarray) -> None:
        x = AxesArray(np.asarray(x), {"ax_sample": 0, "ax_coord": 1})
        n_samples = x.shape[x.ax_sample]
        if self.bagging and self.n_subset is None:
            if self.replace:
                self.n_subset = n_samples
            else:
                self.n_subset = int(0.6 * n_samples)
        if self.bagging and self.n_subset > n_samples and not self.replace:
            warnings.warn(
                "n_subset is larger than sample count without replacement; cannot bag."
            )
            n_subset = n_samples
        else:
            n_subset = self.n_subset

        n_features = x.n_coord
        if self.library_ensemble and self.n_candidates_to_drop > n_features:
            warnings.warn(
                "n_candidates_to_drop larger than number of features.  Cannot "
                "ensemble library."
            )
            n_candidates_to_drop = 0
        else:
            n_candidates_to_drop = self.n_candidates_to_drop

        for _ in range(self.n_models):
            if self.bagging:
                x_ensemble, y_ensemble = _drop_random_samples(
                    x, y, n_subset, self.replace
                )
            else:
                x_ensemble, y_ensemble = x, y

            keep_inds = np.arange(n_features)
            if self.library_ensemble:
                keep_inds = np.sort(
                    np.random.choice(
                        range(n_features),
                        n_features - n_candidates_to_drop,
                        replace=False,
                    )
                )
                x_ensemble = x_ensemble.take(keep_inds, axis=x.ax_coord)
            self.opt.fit(x_ensemble, y_ensemble)
            new_coefs = np.zeros((y.shape[1], n_features))
            new_coefs[:, keep_inds] = self.opt.coef_
            self.coef_list.append(new_coefs)
        # Get average coefficients
        if self.ensemble_aggregator is None:
            self.coef_ = np.median(self.coef_list, axis=0)
        else:
            self.coef_ = self.ensemble_aggregator(self.coef_list)


def _drop_random_samples(
    x: AxesArray,
    x_dot: AxesArray,
    n_subset: int,
    replace: bool,
) -> Tuple[AxesArray]:
    n_samples = x.shape[x.ax_sample]
    rand_inds = np.random.choice(range(n_samples), n_subset, replace=replace)
    x_new = np.take(x, rand_inds, axis=x.ax_sample)
    x_dot_new = np.take(x_dot, rand_inds, axis=x.ax_sample)

    return x_new, x_dot_new


def _normalize_features(x: Float2D) -> Float2D:
    "Calculate the length of vectors and normalize them"
    lengths = np.linalg.norm(x, 2, axis=0)
    return lengths, x / lengths



import warnings
from itertools import combinations
from itertools import combinations_with_replacement as combinations_w_r

from numpy import empty
from numpy import hstack
from numpy import nan_to_num
from numpy import ones
from sklearn.utils.validation import check_is_fitted

from ..utils import AxesArray
from .base import BaseFeatureLibrary
from .base import x_sequence_or_item
from pysindy.differentiation import FiniteDifference


class SINDyPILibrary(BaseFeatureLibrary):
    """
    WARNING: This library is deprecated in PySINDy versions > 1.7. Please
    use the PDE or WeakPDE libraries instead.

    Generate a library with custom functions. The Library takes custom
    libraries for X and Xdot respectively, and then tensor-products them
    together. For a 3D system, a library of constant and linear terms in x_dot,
    i.e. [1, x_dot0, ..., x_dot3], is good
    enough for most problems and implicit terms. The function names list
    should include both X and Xdot functions, without the mixed terms.

    Parameters
    ----------
    library_functions : list of mathematical functions
        Functions to include in the library. Each function will be
        applied to each input variable x.

    x_dot_library_functions : list of mathematical functions
        Functions to include in the library. Each function will be
        applied to each input variable x_dot.

    t : np.ndarray of time slices
        Time base to compute Xdot from X for the implicit terms

    differentiation_method : differentiation object, optional
        Method for differentiating the data. This must be a class extending
        :class:`pysindy.differentiation_methods.base.BaseDifferentiation` class.
        The default option is centered difference.

    function_names : list of functions, optional (default None)
        List of functions used to generate feature names for each library
        function. Each name function must take a string input (representing
        a variable name), and output a string depiction of the respective
        mathematical function applied to that variable. For example, if the
        first library function is sine, the name function might return
        :math:`\\sin(x)` given :math:`x` as input. The function_names
        list must be the same length as library_functions. If no list of
        function names is provided, defaults to using
        :math:`[ f_0(x),f_1(x), f_2(x), \\ldots ]`. For SINDy-PI,
        function_names should include the names of the functions in both the
        x and x_dot libraries (library_functions and x_dot_library_functions),
        but not the mixed terms, which are computed in the code.

    interaction_only : boolean, optional (default True)
        Whether to omit self-interaction terms.
        If True, function evaulations of the form :math:`f(x,x)` and :math:`f(x,y,x)`
        will be omitted, but those of the form :math:`f(x,y)` and :math:`f(x,y,z)`
        will be included.
        If False, all combinations will be included.

    include_bias : boolean, optional (default False)
        If True (default), then include a bias column, the feature in which
        all polynomial powers are zero (i.e. a column of ones - acts as an
        intercept term in a linear model).
        This is hard to do with just lambda functions, because if the system
        is not 1D, lambdas will generate duplicates.

    Attributes
    ----------
    functions : list of functions
        Mathematical library functions to be applied to each input feature.

    function_names : list of functions
        Functions for generating string representations of each library
        function.

    n_features_in_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. The number of output features
        is the product of the number of library functions and the number of
        input features.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.feature_library import SINDyPILibrary
    >>> t = np.linspace(0, 1, 5)
    >>> x = np.ones((5, 2))
    >>> functions = [lambda x: 1, lambda x : np.exp(x),
                     lambda x,y : np.sin(x+y)]
    >>> x_dot_functions = [lambda x: 1, lambda x : x]
    >>> function_names = [lambda x: '',
                          lambda x : 'exp(' + x + ')',
                          lambda x, y : 'sin(' + x + y + ')',
                          lambda x: '',
                  lambda x : x]
    >>> lib = ps.SINDyPILibrary(library_functions=functions,
                                x_dot_library_functions=x_dot_functions,
                                function_names=function_names, t=t
                                ).fit(x)
    >>> lib.transform(x)
            [[ 1.00000000e+00  2.71828183e+00  2.71828183e+00  9.09297427e-01
               2.22044605e-16  6.03579815e-16  6.03579815e-16  2.01904588e-16
               2.22044605e-16  6.03579815e-16  6.03579815e-16  2.01904588e-16]
             [ 1.00000000e+00  2.71828183e+00  2.71828183e+00  9.09297427e-01
               0.00000000e+00  0.00000000e+00  0.00000000e+00  0.00000000e+00
               0.00000000e+00  0.00000000e+00  0.00000000e+00  0.00000000e+00]
             [ 1.00000000e+00  2.71828183e+00  2.71828183e+00  9.09297427e-01
               0.00000000e+00  0.00000000e+00  0.00000000e+00  0.00000000e+00
               0.00000000e+00  0.00000000e+00  0.00000000e+00  0.00000000e+00]
             [ 1.00000000e+00  2.71828183e+00  2.71828183e+00  9.09297427e-01
               0.00000000e+00  0.00000000e+00  0.00000000e+00  0.00000000e+00
               0.00000000e+00  0.00000000e+00  0.00000000e+00  0.00000000e+00]
             [ 1.00000000e+00  2.71828183e+00  2.71828183e+00  9.09297427e-01
              -2.22044605e-16 -6.03579815e-16 -6.03579815e-16 -2.01904588e-16
              -2.22044605e-16 -6.03579815e-16 -6.03579815e-16 -2.01904588e-16]]
    >>> lib.get_feature_names()
        ['', 'exp(x0)', 'exp(x1)', 'sin(x0x1)', 'x0_dot', 'exp(x0)x0_dot',
         'exp(x1)x0_dot', 'sin(x0x1)x0_dot', 'x1_dot', 'exp(x0)x1_dot',
         'exp(x1)x1_dot', 'sin(x0x1)x1_dot']
    """

    def __init__(
        self,
        library_functions=None,
        t=None,
        x_dot_library_functions=None,
        function_names=None,
        interaction_only=True,
        differentiation_method=None,
        include_bias=False,
    ):
        self.x_functions = library_functions
        self.x_dot_functions = x_dot_library_functions
        self.function_names = function_names

        warnings.warn(
            "This library is deprecated in PySINDy versions > 1.7. Please "
            "use the PDE or WeakPDE libraries instead. "
        )
        if library_functions is None and x_dot_library_functions is None:
            raise ValueError(
                "At least one function library, either for x or x_dot, " "is required."
            )
        if x_dot_library_functions is not None and t is None:
            raise ValueError(
                "If using a library that contains x_dot terms,"
                " you must specify a timebase t"
            )
        if function_names is not None:
            x_library_len = 0 if library_functions is None else len(library_functions)
            x_dot_library_len = (
                0 if x_dot_library_functions is None else len(x_dot_library_functions)
            )
            if x_library_len + x_dot_library_len != len(function_names):
                raise ValueError(
                    "(x_library_functions + x_dot_library_functions) and "
                    " function_names must have the same"
                    " number of elements"
                )
        if differentiation_method is None and x_dot_library_functions is not None:
            differentiation_method = FiniteDifference(drop_endpoints=False)
        self.differentiation_method = differentiation_method
        self.interaction_only = interaction_only
        self.t = t
        self.include_bias = include_bias

    @staticmethod
    def _combinations(n_features, n_args, interaction_only):
        """Get the combinations of features to be passed to a library function."""
        comb = combinations if interaction_only else combinations_w_r
        return comb(range(n_features), n_args)

    def get_feature_names(self, input_features=None):
        """Return feature names for output features.

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        check_is_fitted(self)
        n_input_features = self.n_features_in_
        if input_features is None:
            input_features = ["x%d" % i for i in range(n_input_features)]
            x_dot_features = ["x%d_dot" % i for i in range(n_input_features)]
        else:
            x_dot_features = [
                input_features[i] + "_dot" for i in range(n_input_features)
            ]

        feature_names = []
        if self.include_bias:
            feature_names.append("1")

        # Put in normal library for x
        if self.x_functions is not None:
            funcs = self.x_functions
            for i, f in enumerate(funcs):
                for c in self._combinations(
                    n_input_features,
                    f.__code__.co_argcount,
                    self.interaction_only,
                ):
                    feature_names.append(
                        self.function_names[i](*[input_features[j] for j in c])
                    )

        # Put in normal library for x_dot
        if self.x_dot_functions is not None:
            funcs = self.x_dot_functions
            for i, f in enumerate(funcs):
                for c in self._combinations(
                    n_input_features,
                    f.__code__.co_argcount,
                    self.interaction_only,
                ):
                    feature_names.append(
                        self.function_names[-1 - i](*[x_dot_features[j] for j in c])
                    )

        # Put in all the mixed terms
        if self.x_dot_functions is not None and self.x_functions is not None:
            for k, f_dot in enumerate(self.x_dot_functions):
                for f_dot_combs in self._combinations(
                    n_input_features,
                    f_dot.__code__.co_argcount,
                    self.interaction_only,
                ):
                    for i, f in enumerate(self.x_functions):
                        for f_combs in self._combinations(
                            n_input_features,
                            f.__code__.co_argcount,
                            self.interaction_only,
                        ):
                            feature_names.append(
                                self.function_names[i](
                                    *[input_features[comb] for comb in f_combs]
                                )
                                + self.function_names[-1 - k](
                                    *[x_dot_features[comb] for comb in f_dot_combs]
                                )
                            )

        return feature_names

    @x_sequence_or_item
    def fit(self, x_full, y=None):
        """Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            Measurement data.

        Returns
        -------
        self : instance
        """
        n_features = x_full[0].shape[x_full[0].ax_coord]

        self.n_features_in_ = n_features
        n_x_output_features = 0
        n_x_dot_output_features = 0

        # Put in normal x library
        if self.x_functions is not None:
            funcs = self.x_functions
            for f in funcs:
                n_args = f.__code__.co_argcount
                n_x_output_features += len(
                    list(self._combinations(n_features, n_args, self.interaction_only))
                )
            self.n_output_features_ = n_x_output_features

        # Put in normal x_dot library
        if self.x_dot_functions is not None:
            funcs = self.x_dot_functions
            for f in funcs:
                n_args = f.__code__.co_argcount
                n_x_dot_output_features += len(
                    list(self._combinations(n_features, n_args, self.interaction_only))
                )
            self.n_output_features_ += n_x_dot_output_features
            if n_x_output_features != 0:
                self.n_output_features_ += n_x_output_features * n_x_dot_output_features

        if self.function_names is None:
            self.function_names = list(
                map(
                    lambda i: (lambda *x: "f" + str(i) + "(" + ",".join(x) + ")"),
                    range(len(self.x_functions)),
                )
            )
            self.function_names = hstack(
                (
                    self.function_names,
                    list(
                        map(
                            lambda i: (
                                lambda *x_dot: "f_dot"
                                + str(i)
                                + "("
                                + ",".join(x_dot)
                                + ")"
                            ),
                            range(len(self.x_dot_functions)),
                        )
                    ),
                )
            )
        if self.include_bias:
            self.n_output_features_ += 1
        return self

    @x_sequence_or_item
    def transform(self, x_full):
        """Transform data to custom features

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray, shape (n_samples, n_output_features)
            The matrix of features, where n_output_features is the number of features
            generated from applying the custom functions to the inputs.
        """
        check_is_fitted(self)

        xp_full = []
        for x in x_full:
            if self.x_dot_functions is not None:
                x_dot = nan_to_num(self.differentiation_method(x, self.t))

            n_samples, n_features = x.shape

            n_input_features = self.n_features_in_
            if n_features != n_input_features:
                raise ValueError("x shape does not match training shape")

            xp = empty((n_samples, self.n_output_features_), dtype=x.dtype)
            library_idx = 0

            # Put in column of ones in the library
            if self.include_bias:
                xp[:, library_idx] = ones(n_samples)
                library_idx += 1

            # Put in normal x library
            if self.x_functions is not None:
                for i, f in enumerate(self.x_functions):
                    for c in self._combinations(
                        n_input_features,
                        f.__code__.co_argcount,
                        self.interaction_only,
                    ):
                        xp[:, library_idx] = f(*[x[:, j] for j in c])
                        library_idx += 1

            # Put in normal x_dot library
            if self.x_dot_functions is not None:
                for i, f in enumerate(self.x_dot_functions):
                    for c in self._combinations(
                        n_input_features,
                        f.__code__.co_argcount,
                        self.interaction_only,
                    ):
                        xp[:, library_idx] = f(*[x_dot[:, j] for j in c])
                        library_idx += 1

            # Put in mixed x, x_dot terms
            if self.x_dot_functions is not None and self.x_functions is not None:
                for k, f_dot in enumerate(self.x_dot_functions):
                    for f_dot_combs in self._combinations(
                        n_input_features,
                        f_dot.__code__.co_argcount,
                        self.interaction_only,
                    ):
                        for i, f in enumerate(self.x_functions):
                            for f_combs in self._combinations(
                                n_input_features,
                                f.__code__.co_argcount,
                                self.interaction_only,
                            ):
                                xp[:, library_idx] = f(
                                    *[x[:, comb] for comb in f_combs]
                                ) * f_dot(*[x_dot[:, comb] for comb in f_dot_combs])
                                library_idx += 1
            xp_full = xp_full + [AxesArray(xp, x.axes)]
        return xp_full



import warnings
from typing import Union

import numpy as np
from scipy.linalg import LinAlgWarning
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import ridge_regression
from sklearn.utils.validation import check_is_fitted

from .base import BaseOptimizer


class STLSQ(BaseOptimizer):
    """Sequentially thresholded least squares algorithm.
    Defaults to doing Sequentially thresholded Ridge regression.

    Attempts to minimize the objective function
    :math:`\\|y - Xw\\|^2_2 + \\alpha \\|w\\|^2_2`
    by iteratively performing least squares and masking out
    elements of the weight array w that are below a given threshold.

    See the following reference for more details:

        Brunton, Steven L., Joshua L. Proctor, and J. Nathan Kutz.
        "Discovering governing equations from data by sparse
        identification of nonlinear dynamical systems."
        Proceedings of the national academy of sciences
        113.15 (2016): 3932-3937.

    Parameters
    ----------
    threshold : float, optional (default 0.1)
        Minimum magnitude for a coefficient in the weight vector.
        Coefficients with magnitude below the threshold are set
        to zero.

    alpha : float, optional (default 0.05)
        Optional L2 (ridge) regularization on the weight vector.

    max_iter : int, optional (default 20)
        Maximum iterations of the optimization algorithm.

    ridge_kw : dict, optional (default None)
        Optional keyword arguments to pass to the ridge regression.

    normalize_columns : boolean, optional (default False)
        Normalize the columns of x (the SINDy library terms) before regression
        by dividing by the L2-norm. Note that the 'normalize' option in sklearn
        is deprecated in sklearn versions >= 1.0 and will be removed.

    copy_X : boolean, optional (default True)
        If True, X will be copied; else, it may be overwritten.

    initial_guess : np.ndarray, shape (n_features) or (n_targets, n_features),
            optional (default None)
        Initial guess for coefficients ``coef_``.
        If None, least-squares is used to obtain an initial guess.

    verbose : bool, optional (default False)
        If True, prints out the different error terms every iteration.

    sparse_ind : list, optional (default None)
        Indices to threshold and perform ridge regression upon.
        If None, sparse thresholding and ridge regression is applied to all
        indices.

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s).

    ind_ : array, shape (n_features,) or (n_targets, n_features)
        Array of 0s and 1s indicating which coefficients of the
        weight vector have not been masked out, i.e. the support of
        ``self.coef_``.

    history_ : list
        History of ``coef_``. ``history_[k]`` contains the values of
        ``coef_`` at iteration k of sequentially thresholded least-squares.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.integrate import odeint
    >>> from pysindy import SINDy
    >>> from pysindy.optimizers import STLSQ
    >>> lorenz = lambda z,t : [10*(z[1] - z[0]),
    >>>                        z[0]*(28 - z[2]) - z[1],
    >>>                        z[0]*z[1] - 8/3*z[2]]
    >>> t = np.arange(0,2,.002)
    >>> x = odeint(lorenz, [-8,8,27], t)
    >>> opt = STLSQ(threshold=.1, alpha=.5)
    >>> model = SINDy(optimizer=opt)
    >>> model.fit(x, t=t[1]-t[0])
    >>> model.print()
    x0' = -9.999 1 + 9.999 x0
    x1' = 27.984 1 + -0.996 x0 + -1.000 1 x1
    x2' = -2.666 x1 + 1.000 1 x0
    """

    def __init__(
        self,
        threshold=0.1,
        alpha=0.05,
        max_iter=20,
        ridge_kw=None,
        normalize_columns=False,
        copy_X=True,
        initial_guess=None,
        verbose=False,
        sparse_ind=None,
        unbias=True,
    ):
        super().__init__(
            max_iter=max_iter,
            copy_X=copy_X,
            normalize_columns=normalize_columns,
            unbias=unbias,
        )

        if threshold < 0:
            raise ValueError("threshold cannot be negative")
        if alpha < 0:
            raise ValueError("alpha cannot be negative")

        self.threshold = threshold
        self.alpha = alpha
        self.ridge_kw = ridge_kw
        self.initial_guess = initial_guess
        self.verbose = verbose
        self.sparse_ind = sparse_ind

    def _sparse_coefficients(
        self, dim: int, ind_nonzero: np.ndarray, coef: np.ndarray, threshold: float
    ) -> (np.ndarray, np.ndarray):
        """Perform thresholding of the weight vector(s) (on specific indices
        if ``self.sparse_ind`` is not None)"""
        c = np.zeros(dim)
        c[ind_nonzero] = coef
        big_ind = np.abs(c) >= threshold
        if self.sparse_ind is not None:
            nonsparse_ind_mask = np.ones_like(ind_nonzero)
            nonsparse_ind_mask[self.sparse_ind] = False
            big_ind = big_ind | nonsparse_ind_mask
        c[~big_ind] = 0
        return c, big_ind

    def _regress(self, x: np.ndarray, y: np.ndarray, dim: int, sparse_sub: np.ndarray):
        """Perform the ridge regression (on specific indices if
        ``self.sparse_ind`` is not None)"""
        kw = self.ridge_kw or {}
        if self.sparse_ind is None:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=LinAlgWarning)
                try:
                    coef = ridge_regression(x, y, self.alpha, **kw)
                except LinAlgWarning:
                    # increase alpha until warning stops
                    self.alpha = 2 * self.alpha
            self.iters += 1
            return coef
        if self.sparse_ind is not None:
            alpha_array = np.zeros((dim, dim))
            alpha_array[sparse_sub, sparse_sub] = np.sqrt(self.alpha)
            x_lin = np.concatenate((x, alpha_array), axis=0)
            y_lin = np.concatenate((y, np.zeros((dim,))))
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=LinAlgWarning)
                try:
                    coef = (
                        LinearRegression(fit_intercept=False, **kw)
                        .fit(x_lin, y_lin)
                        .coef_
                    )
                except LinAlgWarning:
                    # increase alpha until warning stops
                    self.alpha = 2 * self.alpha
            self.iters += 1
            return coef

    def _no_change(self):
        """Check if the coefficient mask has changed after thresholding"""
        this_coef = self.history_[-1].flatten()
        if len(self.history_) > 1:
            last_coef = self.history_[-2].flatten()
        else:
            last_coef = np.zeros_like(this_coef)
        return all(bool(i) == bool(j) for i, j in zip(this_coef, last_coef))

    def _reduce(self, x, y):
        """Performs at most ``self.max_iter`` iterations of the
        sequentially-thresholded least squares algorithm.

        Assumes an initial guess for coefficients and support are saved in
        ``self.coef_`` and ``self.ind_``.
        """
        if self.initial_guess is not None:
            self.coef_ = self.initial_guess

        ind = self.ind_
        n_samples, n_features = x.shape
        n_targets = y.shape[1]
        n_features_selected = np.sum(ind)
        sparse_sub = [np.array(self.sparse_ind)] * y.shape[1]
        optvar = np.zeros((n_targets, n_features))

        # Print initial values for each term in the optimization
        if self.verbose:
            row = [
                "Iteration",
                "|y - Xw|^2",
                "a * |w|_2",
                "|w|_0",
                "Total error: |y - Xw|^2 + a * |w|_2",
            ]
            print(
                "{: >10} ... {: >10} ... {: >10} ... {: >10}"
                " ... {: >10}".format(*row)
            )

        for k in range(self.max_iter):
            if np.count_nonzero(ind) == 0:
                warnings.warn(
                    "Sparsity parameter is too big ({}) and eliminated all "
                    "coefficients".format(self.threshold)
                )
                optvar = np.zeros((n_targets, n_features))
                break

            optvar = np.zeros((n_targets, n_features))
            for i in range(n_targets):
                if np.count_nonzero(ind[i]) == 0:
                    warnings.warn(
                        "Sparsity parameter is too big ({}) and eliminated all "
                        "coefficients".format(self.threshold)
                    )
                    continue
                coef_i = self._regress(
                    x[:, ind[i]], y[:, i], np.count_nonzero(ind[i]), sparse_sub[i]
                )
                coef_i, ind_i = self._sparse_coefficients(
                    n_features, ind[i], coef_i, self.threshold
                )
                if self.sparse_ind is not None:
                    vals_to_remove = np.intersect1d(
                        self.sparse_ind, np.where(coef_i == 0)
                    )
                    sparse_sub[i] = _remove_and_decrement(
                        self.sparse_ind, vals_to_remove
                    )
                optvar[i] = coef_i
                ind[i] = ind_i

            self.history_.append(optvar)
            if self.verbose:
                R2 = np.sum((y - np.dot(x, optvar.T)) ** 2)
                L2 = self.alpha * np.sum(optvar**2)
                L0 = np.count_nonzero(optvar)
                row = [k, R2, L2, L0, R2 + L2]
                print(
                    "{0:10d} ... {1:10.4e} ... {2:10.4e} ... {3:10d}"
                    " ... {4:10.4e}".format(*row)
                )
            if np.sum(ind) == n_features_selected or self._no_change():
                # could not (further) select important features
                break
        else:
            warnings.warn(
                "STLSQ did not converge after {self.max_iter} iterations.",
                ConvergenceWarning,
            )
        if self.sparse_ind is None:
            self.coef_ = optvar
            self.ind_ = ind
        else:
            non_sparse_ind = np.setxor1d(self.sparse_ind, range(n_features))
            self.coef_ = optvar[:, self.sparse_ind]
            self.ind_ = ind[:, self.sparse_ind]
            self.optvar_non_sparse_ = optvar[:, non_sparse_ind]
            self.ind_non_sparse_ = ind[:, non_sparse_ind]

    @property
    def complexity(self):
        check_is_fitted(self)

        return np.count_nonzero(self.coef_) + np.count_nonzero(
            [abs(self.intercept_) >= self.threshold]
        )

    def _unbias(self, x: np.ndarray, y: np.ndarray):
        if not self.sparse_ind:
            return super()._unbias(x, y)
        regression_col_mask = np.zeros((y.shape[1], x.shape[1]), dtype=bool)
        regression_col_mask[:, self.sparse_ind] = self.ind_
        non_sparse_ind = np.setxor1d(self.sparse_ind, range(x.shape[1]))
        regression_col_mask[:, non_sparse_ind] = self.ind_non_sparse_

        for i in range(self.ind_.shape[0]):
            if np.any(self.ind_[i]):
                optvar = (
                    LinearRegression(fit_intercept=False)
                    .fit(x[:, regression_col_mask[i]], y[:, i])
                    .coef_
                )
                self.coef_[i] = optvar[self.sparse_ind]
                self.optvar_non_sparse_[i] = optvar[non_sparse_ind]


def _remove_and_decrement(
    existing_vals: Union[np.ndarray, list], vals_to_remove: Union[np.ndarray, list]
) -> np.ndarray:
    """Remove elements from existing values and decrement the elements
    that are greater than the removed elements"""
    for s in reversed(vals_to_remove):
        existing_vals = np.delete(existing_vals, np.where(s == existing_vals))
        existing_vals = np.where(existing_vals > s, existing_vals - 1, existing_vals)
    return existing_vals



"""
Base class for numerical differentiation methods
"""
import abc

from sklearn.base import BaseEstimator


class BaseDifferentiation(BaseEstimator):
    """
    Base class for differentiation methods.

    Simply forces differentiation methods to implement a
    ``_differentiate`` function.

    Attributes:
        smoothed_x_: Methods that smooth x before differentiating save
            that value here.  Methods that do not simply save x here.
    """

    def __init__(self):
        pass

    # Force subclasses to implement this
    @abc.abstractmethod
    def _differentiate(self, x, t=1):
        """
        Numerically differentiate data.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_input_features)
            Data to be differentiated. Rows of x should correspond to the same
            point in time.

        t: float or numpy array of shape (n_samples,)
            If t is a float, it is interpreted as the timestep between
            samples in x.
            If t is a numpy array, it specifies the times corresponding
            to the rows of x. That is, t[i] should be the time at which
            the measurements x[i, :] were taken.
            The points in t are assumed to be increasing.

        Returns
        -------
        x_dot: array-like, shape (n_samples, n_input_features)
            Numerical time derivative of x. Entries where derivatives were
            not computed will have the value np.nan.
        """
        raise NotImplementedError

    def __call__(self, x, t=1):
        return self._differentiate(x, t)



from .base import BaseOptimizer
from .base import EnsembleOptimizer
from .frols import FROLS
from .sr3 import SR3
from .ssr import SSR
from .stlsq import STLSQ
from .wrapped_optimizer import WrappedOptimizer

try:
    from ._constrained_sr3 import ConstrainedSR3
except (ImportError, NameError):
    pass
try:
    from .miosr import MIOSR
except (ImportError, NameError):
    pass
try:
    from .trapping_sr3 import TrappingSR3
except (ImportError, NameError):
    pass
try:
    from .sindy_pi import SINDyPI
except (ImportError, NameError):
    pass
try:
    from .stable_linear_sr3 import StableLinearSR3
except (ImportError, NameError):
    pass
try:
    from .sbr import SBR
except (ImportError, NameError):
    pass


__all__ = [
    "BaseOptimizer",
    "EnsembleOptimizer",
    "WrappedOptimizer",
    "SR3",
    "STLSQ",
    "ConstrainedSR3",
    "StableLinearSR3",
    "TrappingSR3",
    "SSR",
    "FROLS",
    "SINDyPI",
    "MIOSR",
    "SBR",
]



try:
    import gurobipy as gp
except ImportError:
    raise ImportError(
        "To use MIOSR please install pysindy with pip install pysindy[miosr]"
        "to gain access to a restricted installation of Gurobi."
        "Free unrestricted academic licenses are available at "
        "https://www.gurobi.com/academia/academic-program-and-licenses/"
    )

import numpy as np
from sklearn.utils.validation import check_is_fitted

from .base import BaseOptimizer


class MIOSR(BaseOptimizer):
    """Mixed-Integer Optimized Sparse Regression.

    Solves the sparsity constrained regression problem to provable optimality
    .. math::

        \\|y-Xw\\|^2_2 + \\lambda R(u)

    .. math::

        \\text{subject to } \\|w\\|_0 \\leq k

    by using type-1 specially ordered sets (SOS1) to encode the support of
    the coefficients. Can optionally add additional constraints on the
    coefficients or access the gurobi model directly for advanced usage.
    See the following reference for additional details:

        Bertsimas, D. and Gurnee, W., 2022. Learning Sparse Nonlinear Dynamics
        via Mixed-Integer Optimization. arXiv preprint arXiv:2206.00176.

    Parameters
    ----------
    target_sparsity : int, optional (default 5)
        The maximum number of nonzero coefficients across all dimensions.
        If set, the model will fit all dimensions jointly, potentially reducing
        statistical efficiency.

    group_sparsity : int tuple, optional (default None)
        Tuple of length n_targets constraining the number of nonzero
        coefficients for each target dimension.

    alpha : float, optional (default 0.01)
        Optional L2 (ridge) regularization on the weight vector.

    regression_timeout : int, optional (default 10)
        The timeout (in seconds) of the gurobi optimizer to solve and prove
        optimality (either per dimension or jointly depending on the
        above sparsity settings).

    constraint_lhs : numpy ndarray, optional (default None)
        Shape should be (n_constraints, n_features * n_targets),
        The left hand side matrix C of Cw <= d.
        There should be one row per constraint.

    constraint_rhs : numpy ndarray, shape (n_constraints,), optional (default None)
        The right hand side vector d of Cw <= d.

    constraint_order : string, optional (default "target")
        The format in which the constraints ``constraint_lhs`` were passed.
        Must be one of "target" or "feature".
        "target" indicates that the constraints are grouped by target:
        i.e. the first ``n_features`` columns
        correspond to constraint coefficients on the library features
        for the first target (variable), the next ``n_features`` columns to
        the library features for the second target (variable), and so on.
        "feature" indicates that the constraints are grouped by library
        feature: the first ``n_targets`` columns correspond to the first
        library feature, the next ``n_targets`` columns to the second library
        feature, and so on.

    normalize_columns : boolean, optional (default False)
        Normalize the columns of x (the SINDy library terms) before regression
        by dividing by the L2-norm. Note that the 'normalize' option in sklearn
        is deprecated in sklearn versions >= 1.0 and will be removed. Note that
        this parameter is incompatible with the constraints!

    copy_X : boolean, optional (default True)
        If True, X will be copied; else, it may be overwritten.

    initial_guess : np.ndarray, shape (n_features) or (n_targets, n_features), \
            optional (default None)
        Initial guess for coefficients ``coef_`` to warmstart the optimizer.

    verbose : bool, optional (default False)
        If True, prints out the Gurobi solver log.

    unbias: bool
        Required to be false, maintained for supertype compatibility

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s).
    ind_ : array, shape (n_features,) or (n_targets, n_features)
        Array of 0s and 1s indicating which coefficients of the
        weight vector have not been masked out, i.e. the support of
        ``self.coef_``.
    model : gurobipy.model
        The raw gurobi model being solved.
    """

    def __init__(
        self,
        target_sparsity=5,
        group_sparsity=None,
        alpha=0.01,
        regression_timeout=10,
        constraint_lhs=None,
        constraint_rhs=None,
        constraint_order="target",
        normalize_columns=False,
        copy_X=True,
        initial_guess=None,
        verbose=False,
        unbias=False,
    ):
        super().__init__(
            normalize_columns=normalize_columns,
            copy_X=copy_X,
            unbias=unbias,
        )

        if target_sparsity is not None and (
            target_sparsity <= 0 or not isinstance(target_sparsity, int)
        ):
            raise ValueError("target_sparsity must be positive int")
        if constraint_order not in {"target", "feature"}:
            raise ValueError("constraint_order must be one of {'target', 'feature'}")
        if alpha < 0:
            raise ValueError("alpha cannot be negative")
        if unbias and constraint_lhs is not None:
            raise ValueError("MIOSR is incompatible with unbiasing. Set unbias=False")
        self.target_sparsity = target_sparsity
        self.group_sparsity = group_sparsity
        self.constraint_lhs = constraint_lhs
        self.constraint_rhs = constraint_rhs
        self.constraint_order = constraint_order
        self.alpha = alpha
        self.initial_guess = initial_guess
        self.regression_timeout = regression_timeout
        self.verbose = verbose

        self.model = None

    def _make_model(self, X, y, k, warm_start=None):
        model = gp.Model()
        n_samples, n_features = X.shape
        _, n_targets = y.shape

        coeff_var = model.addMVar(
            n_targets * n_features,
            lb=-gp.GRB.INFINITY,
            vtype=gp.GRB.CONTINUOUS,
            name="coeff_var",
        )
        iszero = model.addMVar(
            n_targets * n_features, vtype=gp.GRB.BINARY, name="iszero"
        )

        # Sparsity constraint
        for i in range(n_targets * n_features):
            model.addSOS(gp.GRB.SOS_TYPE1, [coeff_var[i], iszero[i]])
        model.addConstr(iszero.sum() >= (n_targets * n_features) - k, name="sparsity")

        # Group sparsity constraints
        if self.group_sparsity is not None and n_targets > 1:
            for i in range(n_targets):
                dimension_sparsity = self.group_sparsity[i]
                model.addConstr(
                    iszero[i * n_features : (i + 1) * n_features].sum()
                    >= n_features - dimension_sparsity,
                    name=f"group_sparsity{i}",
                )

        # General equality constraints
        if self.constraint_lhs is not None and self.constraint_rhs is not None:
            if self.constraint_order == "feature":
                target_indexing = (
                    np.arange(n_targets * n_features)
                    .reshape(n_targets, n_features, order="F")
                    .flatten()
                )
                constraint_lhs = self.constraint_lhs[:, target_indexing]
            else:
                constraint_lhs = self.constraint_lhs
            model.addConstr(
                constraint_lhs @ coeff_var == self.constraint_rhs, name="coeff_constrs"
            )

        if warm_start is not None:
            warm_start = warm_start.reshape(1, n_targets * n_features)[0]
            for i in range(n_features):
                iszero[i].start = abs(warm_start[i]) < 1e-6
                coeff_var[i].start = warm_start[i]

        # Equation 15 in paper
        Quad = np.dot(X.T, X)
        obj = self.alpha * (coeff_var @ coeff_var)
        for i in range(n_targets):
            lin = np.dot(y[:, i].T, X)
            obj += (
                coeff_var[n_features * i : n_features * (i + 1)]
                @ Quad
                @ coeff_var[n_features * i : n_features * (i + 1)]
            )
            obj -= 2 * (lin @ coeff_var[n_features * i : n_features * (i + 1)])

        model.setObjective(obj, gp.GRB.MINIMIZE)

        model.params.OutputFlag = 1 if self.verbose else 0
        model.params.timelimit = self.regression_timeout
        model.update()

        self.model = model

        return model, coeff_var

    def _regress(self, X, y, k, warm_start=None):
        """
        Deploy and optimize the MIO formulation of L0-Regression.
        """
        m, coeff_var = self._make_model(X, y, k, warm_start)
        m.optimize()
        return coeff_var.X

    def _reduce(self, x, y):
        """
        Runs MIOSR either per dimension or jointly on all dimensions.

        Assumes an initial guess for coefficients and support are saved in
        ``self.coef_`` and ``self.ind_``.
        """
        if self.initial_guess is not None:
            self.coef_ = self.initial_guess

        n_samples, n_features = x.shape
        _, n_targets = y.shape

        if (
            self.target_sparsity is not None or self.constraint_lhs is not None
        ):  # Regress jointly
            coefs = self._regress(x, y, self.target_sparsity, self.initial_guess)
            # Remove nonzero terms due to numerical error
            non_active_ixs = np.argsort(np.abs(coefs))[: -int(self.target_sparsity)]
            coefs[non_active_ixs] = 0
            self.coef_ = coefs.reshape(n_targets, n_features)
            self.ind_ = (np.abs(self.coef_) > 1e-6).astype(int)
        else:  # Regress dimensionwise
            for i in range(n_targets):
                k = self.group_sparsity[i]
                warm_start = (
                    None if self.initial_guess is None else self.initial_guess[[i], :]
                )
                coef_i = self._regress(x, y[:, [i]], k, warm_start=warm_start)
                # Remove nonzero terms due to numerical error
                non_active_ixs = np.argsort(np.abs(coef_i))[: -int(k)]
                coef_i[non_active_ixs] = 0
                self.coef_[i, :] = coef_i
            self.ind_ = (np.abs(self.coef_) > 1e-6).astype(int)

    @property
    def complexity(self):
        check_is_fitted(self)
        return np.count_nonzero(self.coef_)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["model"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.model = None



"""
`Deeptime <https://deeptime-ml.github.io/>`_ wrapper interface for PySINDy.
"""
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from ._core import SINDy
from .utils import SampleConcatter


class SINDyEstimator(SINDy):
    """
    Implementation of SINDy conforming to the API of a Deeptime
    `Estimator \
    <https://deeptime-ml.github.io/api/generated/deeptime.base.Estimator.html>`_.

    Parameters
    ----------
    optimizer : optimizer object, optional
        Optimization method used to fit the SINDy model. This must be an object
        extending :class:`pysindy.optimizers.BaseOptimizer`. Default is
        sequentially thresholded least squares with a threshold of 0.1.

    feature_library : feature library object, optional
        Feature library object used to specify candidate right-hand side features.
        This must be an object extending the
        :class:`pysindy.feature_library.base.BaseFeatureLibrary`.
        Default is polynomial features of degree 2.

    differentiation_method : differentiation object, optional
        Method for differentiating the data. This must be an object extending
        the :class:`pysindy.differentiation_methods.base.BaseDifferentiation` class.
        Default is centered difference.

    feature_names : list of string, length n_input_features, optional
        Names for the input features (e.g. ``['x', 'y', 'z']``). If None, will use
        ``['x0', 'x1', ...]``.

    Attributes
    ----------
    model : sklearn.multioutput.MultiOutputRegressor object
        The fitted SINDy model.

    n_input_features_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. This number is a function of
        ``self.n_input_features`` and the feature library being used.

    """

    def __init__(
        self,
        optimizer=None,
        feature_library=None,
        differentiation_method=None,
    ):
        super(SINDyEstimator, self).__init__(
            optimizer=optimizer,
            feature_library=feature_library,
            differentiation_method=differentiation_method,
        )
        self._model = None

    def fit(self, x, **kwargs):
        """
        Fit the SINDyEstimator to data, learning a dynamical systems model
        for the data.

        Parameters
        ----------
        x: array-like or list of array-like, shape (n_samples, n_input_features)
            Training data. If training data contains multiple trajectories,
            x should be a list containing data for each trajectory. Individual
            trajectories may contain different numbers of samples.

        **kwargs: dict, optional
            Optional keyword arguments to pass to :meth:`fit` method.

        Returns
        -------
        self: fitted :class:`SINDyEstimator` instance
        """
        super(SINDyEstimator, self).fit(x, **kwargs)
        self._model = SINDyModel(
            feature_library=self.feature_library,
            optimizer=self.optimizer,
            n_control_features_=self.n_control_features_,
        )
        return self

    def fetch_model(self):
        """
        Yields the estimated model. Can be none if :meth:`fit` was not called.

        Returns
        -------
        model: :class:`SINDyModel` or None
            The estimated SINDy model or none
        """
        return self._model

    @property
    def has_model(self):
        """Property reporting whether this estimator contains an estimated
        model. This assumes that the model is initialized with ``None`` otherwise.

        :type: bool
        """
        return self._model is not None


class SINDyModel(SINDy):
    """
    Implementation of SINDy conforming to the API of a Deeptime
    `Model <https://deeptime-ml.github.io/api/generated/deeptime.base.Model.html>`_.

    The model is represented as a Scikit-learn pipeline object with three steps:
    1. Map the raw input data to nonlinear features according to the selected
    ``feature_library``
    2. Reshape the data from input shape to an optimization problem
    3. Multiply the nonlinear features with a coefficient matrix encapuslated
    in ``optimizer``.

    This class expects the feature library and optimizer to already be fit
    with a :class:`SINDyEstimator`. It is best to instantiate a :class:`SINDyModel`
    object via the :meth:`SINDyEstimator.fetch_model()` rather than calling
    the :class:`SINDyModel` constructor directly.

    Parameters
    ----------
    optimizer : optimizer object
        Optimization method used to fit the SINDy model. This must be an
        (already fit) object extending :class:`pysindy.optimizers.BaseOptimizer`.

    feature_library : feature library object
        Feature library object used to specify candidate right-hand side features.
        This must be an (already fit) object extending
        :class:`pysindy.feature_library.BaseFeatureLibrary`.

    differentiation_method : differentiation object
        Method for differentiating the data. This must be an object extending
        :class:`pysindy.differentiation_methods.BaseDifferentiation`.
        Default is centered difference.

    feature_names : list of string, length n_input_features, optional
        Names for the input features (e.g. ``['x', 'y', 'z']``). If None, will use
        ``['x0', 'x1', ...]``.

    Attributes
    ----------
    model : sklearn.multioutput.MultiOutputRegressor object
        The fitted SINDy model.

    n_input_features_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. This number is a function of
        ``self.n_input_features`` and the feature library being used.
    """

    def __init__(
        self,
        feature_library,
        optimizer,
        n_control_features_=0,
    ):
        super(SINDyModel, self).__init__(
            feature_library=feature_library,
            optimizer=optimizer,
        )
        self.n_control_features_ = n_control_features_

        check_is_fitted(feature_library)
        check_is_fitted(optimizer)

        steps = [
            ("features", feature_library),
            ("shaping", SampleConcatter()),
            ("model", optimizer),
        ]
        self.model = Pipeline(steps)

        self.n_features_in_ = self.model.steps[0][1].n_features_in_
        self.n_output_features_ = self.model.steps[0][1].n_output_features_

    def copy(self):
        """Makes a deep copy of this model.

        Returns
        -------
        copy: :class:`SINDyModel`
            A new copy of this model.
        """
        import copy

        return copy.deepcopy(self)



import warnings
from functools import wraps
from typing import Callable
from typing import Sequence
from typing import Union

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import bisect
from sklearn.base import MultiOutputMixin
from sklearn.utils.validation import check_array

from ._axes import AxesArray

# Define a special object for the default value of t in
# validate_input. Normally we would set the default
# value of t to be None, but it is possible for the user
# to pass in None, in which case validate_input performs
# no checks on t.
T_DEFAULT = object()


def flatten_2d_tall(x):
    return x.reshape(x.size // x.shape[-1], x.shape[-1])


def validate_input(x, t=T_DEFAULT):
    """Forces input data to have compatible dimensions, if possible.

    Args:
        x: array of input data (measured coordinates across time)
        t: time values for measurements.

    Returns:
        x as 2D array, with time dimension on first axis and coordinate
        index on second axis.
    """
    if not isinstance(x, np.ndarray):
        raise ValueError("x must be array-like")
    elif x.ndim == 1:
        x = x.reshape(-1, 1)
    x_new = flatten_2d_tall(x)

    check_array(x, ensure_2d=False, allow_nd=True)

    if t is not T_DEFAULT:
        if t is None:
            raise ValueError("t must be a scalar or array-like.")
        # Apply this check if t is a scalar
        elif np.ndim(t) == 0 and (isinstance(t, int) or isinstance(t, float)):
            if t <= 0:
                raise ValueError("t must be positive")
        # Only apply these tests if t is array-like
        elif isinstance(t, np.ndarray):
            if not len(t) == x.shape[-2]:
                raise ValueError("Length of t should match x.shape[-2].")
            if not np.all(t[:-1] < t[1:]):
                raise ValueError("Values in t should be in strictly increasing order.")
        else:
            raise ValueError("t must be a scalar or array-like.")

    return x_new


def validate_no_reshape(x, t: Union[float, np.ndarray, object] = T_DEFAULT):
    """Check types and numerical sensibility of arguments.

    Args:
        x: array of input data (measured coordinates across time)
        t: time values for measurements.

    Returns:
        x as 2D array, with time dimension on first axis and coordinate
        index on second axis.
    """
    if not hasattr(x, "shape"):
        raise TypeError("Input value must be array-like")
    check_array(x, ensure_2d=False, allow_nd=True)

    if t is not T_DEFAULT:
        if t is None:
            raise ValueError("t must be a scalar or array-like.")
        # Apply this check if t is a scalar
        elif np.ndim(t) == 0 and (isinstance(t, int) or isinstance(t, float)):
            if t <= 0:
                raise ValueError("t must be positive")
        # Only apply these tests if t is array-like
        elif hasattr(t, "shape"):
            if not len(t) == x.shape[-2]:
                raise ValueError("Length of t should match x.shape[-2].")
            if not np.all(t[:-1] < t[1:]):
                raise ValueError("Values in t should be in strictly increasing order.")
        else:
            raise ValueError("t must be a scalar or array-like.")
    return x


def validate_control_variables(x: Sequence[AxesArray], u: Sequence[AxesArray]) -> None:
    """Ensure that control variables u are compatible with the data x.

    Args:
        x: trajectories of system variables
        u: trajectories of control variables
    """
    if not isinstance(x, Sequence):
        raise ValueError("x must be a Sequence")
    if not isinstance(u, Sequence):
        raise ValueError("u must be a Sequence")
    if len(x) != len(u):
        raise ValueError("x and u must be the same length")

    def _check_control_shape(x, u):
        """
        Compare shape of control variable u against x.
        """
        if u.n_time != x.n_time:
            raise ValueError(
                "control variables u must have same number of time points as x. "
                f"u has {u.n_time} time points and x has {x.n_time} time points"
            )
        return u

    u_arr = [_check_control_shape(xi, ui) for xi, ui in zip(x, u)]

    return u_arr


def drop_nan_samples(x, y):
    """Drops samples from x and y where either has a nan value"""
    x_non_sample_axes = tuple(ax for ax in range(x.ndim) if ax != x.ax_sample)
    y_non_sample_axes = tuple(ax for ax in range(y.ndim) if ax != y.ax_sample)
    x_good_samples = (~np.isnan(x)).any(axis=x_non_sample_axes)
    y_good_samples = (~np.isnan(y)).any(axis=y_non_sample_axes)
    good_sample_ind = np.nonzero(x_good_samples & y_good_samples)[0]
    x = x.take(good_sample_ind, axis=x.ax_sample)
    y = y.take(good_sample_ind, axis=y.ax_sample)
    return x, y


def reorder_constraints(arr, n_features, output_order="feature"):
    """Switch between 'feature' and 'target' constraint order."""
    warnings.warn("Target format constraints are deprecated.", stacklevel=2)
    n_constraints = arr.shape[0] if arr.ndim > 1 else 1
    n_tgt = arr.size // n_features // n_constraints
    if output_order == "feature":
        starting_shape = (n_constraints, n_tgt, n_features)
    else:
        starting_shape = (n_constraints, n_features, n_tgt)

    return arr.reshape(starting_shape).transpose([0, 2, 1]).reshape((n_constraints, -1))


def _validate_prox_and_reg_inputs(func):
    """Add guard code to ensure weight and argument have compatible shape/type

    Decorates prox and regularization functions.
    """

    @wraps(func)
    def wrapper(x, regularization_weight):
        if isinstance(regularization_weight, np.ndarray):
            weight_shape = regularization_weight.shape
            if weight_shape != x.shape:
                raise ValueError(
                    f"Invalid shape for 'regularization_weight': "
                    f"{weight_shape}. Must be the same shape as x: {x.shape}."
                )
        elif not isinstance(regularization_weight, (int, float)):
            raise ValueError("'regularization_weight' must be a scalar")
        return func(x, regularization_weight)

    return wrapper


def get_prox(
    regularization: str,
) -> Callable[
    [NDArray[np.float64], Union[float, NDArray[np.float64]]], NDArray[np.float64]
]:
    """
    Args:
    -----
    regularization: 'l0' | 'weighted_l0' | 'l1' | 'weighted_l1' | 'l2' | 'weighted_l2'

    Returns:
    --------
    proximal_operator: (x: np.array, reg_weight: float | np.array) -> np.array
        A function that takes an input array x and a regularization weight,
        which can be either a scalar or array of the same shape,
        and returns an array of the same shape
    """

    prox = {
        "l0": _prox_l0,
        "weighted_l0": _prox_l0,
        "l1": _prox_l1,
        "weighted_l1": _prox_l1,
        "l2": _prox_l2,
        "weighted_l2": _prox_l2,
    }
    regularization = regularization.lower()
    return prox[regularization]


@_validate_prox_and_reg_inputs
def _prox_l0(
    x: NDArray[np.float64],
    regularization_weight: Union[float, NDArray[np.float64]],
):
    threshold = np.sqrt(2 * regularization_weight)
    return x * (np.abs(x) > threshold)


@_validate_prox_and_reg_inputs
def _prox_l1(
    x: NDArray[np.float64],
    regularization_weight: Union[float, NDArray[np.float64]],
):

    return np.sign(x) * np.maximum(np.abs(x) - regularization_weight, 0)


@_validate_prox_and_reg_inputs
def _prox_l2(
    x: NDArray[np.float64],
    regularization_weight: Union[float, NDArray[np.float64]],
):
    return x / (1 + 2 * regularization_weight)


@_validate_prox_and_reg_inputs
def _regularization_l0(
    x: NDArray[np.float64],
    regularization_weight: Union[float, NDArray[np.float64]],
):
    return np.sum(regularization_weight * (x != 0))


@_validate_prox_and_reg_inputs
def _regularization_l1(
    x: NDArray[np.float64],
    regularization_weight: Union[float, NDArray[np.float64]],
):
    return np.sum(regularization_weight * np.abs(x))


@_validate_prox_and_reg_inputs
def _regularization_l2(
    x: NDArray[np.float64],
    regularization_weight: Union[float, NDArray[np.float64]],
):
    return np.sum(regularization_weight * x**2)


def get_regularization(
    regularization: str,
) -> Callable[[NDArray[np.float64], Union[float, NDArray[np.float64]]], float]:
    """
    Args:
    -----
    regularization: 'l0' | 'weighted_l0' | 'l1' | 'weighted_l1' | 'l2' | 'weighted_l2'

    Returns:
    --------
    regularization_function: (x: np.array, reg_weight: float | np.array) -> np.array
        A function that takes an input array x and a regularization weight,
        which can be either a scalar or array of the same shape,
        and returns a float
    """

    regularization_fn = {
        "l0": _regularization_l0,
        "weighted_l0": _regularization_l0,
        "l1": _regularization_l1,
        "weighted_l1": _regularization_l1,
        "l2": _regularization_l2,
        "weighted_l2": _regularization_l2,
    }
    regularization = regularization.lower()
    return regularization_fn[regularization]


def capped_simplex_projection(trimming_array, trimming_fraction):
    """Projection of trimming_array onto the capped simplex"""
    a = np.min(trimming_array) - 1.0
    b = np.max(trimming_array) - 0.0

    def f(x):
        return (
            np.sum(np.maximum(np.minimum(trimming_array - x, 1.0), 0.0))
            - (1.0 - trimming_fraction) * trimming_array.size
        )

    x = bisect(f, a, b)

    return np.maximum(np.minimum(trimming_array - x, 1.0), 0.0)


def supports_multiple_targets(estimator):
    """Checks whether estimator supports multiple targets."""
    if isinstance(estimator, MultiOutputMixin):
        return True
    try:
        return estimator._more_tags()["multioutput"]
    except (AttributeError, KeyError):
        return False



import numpy as np
import pytest
from scipy.integrate import solve_ivp

from pysindy.utils.odes import bacterial
from pysindy.utils.odes import burgers_galerkin
from pysindy.utils.odes import cubic_damped_SHO
from pysindy.utils.odes import cubic_oscillator
from pysindy.utils.odes import double_pendulum
from pysindy.utils.odes import duffing
from pysindy.utils.odes import enzyme
from pysindy.utils.odes import hopf
from pysindy.utils.odes import kinematic_commonroad
from pysindy.utils.odes import linear_3D
from pysindy.utils.odes import linear_damped_SHO
from pysindy.utils.odes import lorenz
from pysindy.utils.odes import lorenz_control
from pysindy.utils.odes import lotka
from pysindy.utils.odes import meanfield
from pysindy.utils.odes import mhd
from pysindy.utils.odes import oscillator
from pysindy.utils.odes import pendulum_on_cart
from pysindy.utils.odes import rossler
from pysindy.utils.odes import van_der_pol
from pysindy.utils.odes import yeast


@pytest.mark.parametrize(
    "ode_params",
    [
        (bacterial, 2),
        (cubic_damped_SHO, 2),
        (cubic_oscillator, 2),
        (double_pendulum, 4),
        (duffing, 2),
        (enzyme, 1),
        (hopf, 2),
        (kinematic_commonroad, 5, 2),
        (linear_3D, 3),
        (linear_damped_SHO, 2),
        (lorenz, 3),
        (lorenz_control, 3, 2),
        (lotka, 2),
        (meanfield, 3),
        (mhd, 6),
        (oscillator, 3),
        (pendulum_on_cart, 4),
        (rossler, 3),
        (van_der_pol, 2),
        (yeast, 7),
    ],
)
def test_odes(ode_params):
    def u_fun(t):
        return np.column_stack([np.sin(2 * t), t**2])

    t = np.linspace(0, 10, 100)
    x0 = np.random.rand(ode_params[1])
    if np.shape(ode_params)[0] == 3:
        x_sim = solve_ivp(ode_params[0], (t[0], t[-1]), x0, t_eval=t, args=(u_fun,)).y.T
    else:
        x_sim = solve_ivp(ode_params[0], (t[0], t[-1]), x0, t_eval=t).y.T
    assert np.max(abs(x_sim)) <= 1e5  # avoided unbounded growth


# define analytic galerkin model for quadratic nonlinear systems
def galerkin_model(a, L, Q):
    """RHS of POD-Galerkin model, for time integration"""
    return (L @ a) + np.einsum("ijk,j,k->i", Q, a, a)


def test_galerkin_models():
    # get analytic L and Q operators and galerkin model
    L, Q = burgers_galerkin()

    def rhs(t, a):
        return galerkin_model(a, L, Q)

    # Generate initial condition from unstable eigenvectors
    lamb, Phi = np.linalg.eig(L)
    idx = np.argsort(-np.real(lamb))
    lamb, Phi = lamb[idx], Phi[:, idx]
    a0 = np.real(1e-4 * Phi[:, :2] @ np.random.random((2)))

    # define parameters
    dt = 1e-3
    t_sim = np.arange(0, 300, dt)
    x_sim = solve_ivp(rhs, (t_sim[0], t_sim[-1]), a0, t_eval=t_sim).y.T
    assert np.max(abs(x_sim)) <= 1e5  # avoided unbounded growth



from ._axes import AxesArray
from ._axes import comprehend_axes
from ._axes import concat_sample_axis
from ._axes import SampleConcatter
from ._axes import wrap_axes
from .base import capped_simplex_projection
from .base import drop_nan_samples
from .base import flatten_2d_tall
from .base import get_prox
from .base import get_regularization
from .base import reorder_constraints
from .base import supports_multiple_targets
from .base import validate_control_variables
from .base import validate_input
from .base import validate_no_reshape
from .odes import bacterial
from .odes import burgers_galerkin
from .odes import cubic_damped_SHO
from .odes import cubic_oscillator
from .odes import double_pendulum
from .odes import duffing
from .odes import enzyme
from .odes import hopf
from .odes import kinematic_commonroad
from .odes import linear_3D
from .odes import linear_damped_SHO
from .odes import logistic_map
from .odes import logistic_map_control
from .odes import logistic_map_multicontrol
from .odes import lorenz
from .odes import lorenz_control
from .odes import lorenz_u
from .odes import lotka
from .odes import meanfield
from .odes import mhd
from .odes import oscillator
from .odes import pendulum_on_cart
from .odes import rossler
from .odes import van_der_pol
from .odes import yeast

# from .base import convert_u_dot_integral
# from .base import integrate
# from .base import integrate2
# from .base import phi
# from .base import linear_weights

__all__ = [
    "AxesArray",
    "SampleConcatter",
    "concat_sample_axis",
    "wrap_axes",
    "comprehend_axes",
    "capped_simplex_projection",
    "drop_nan_samples",
    "get_prox",
    "get_regularization",
    "reorder_constraints",
    "supports_multiple_targets",
    "validate_control_variables",
    "validate_input",
    "validate_no_reshape",
    "flatten_2d_tall",
    "linear_damped_SHO",
    "cubic_damped_SHO",
    "linear_3D",
    "lotka",
    "van_der_pol",
    "duffing",
    "rossler",
    "cubic_oscillator",
    "hopf",
    "lorenz",
    "lorenz_control",
    "lorenz_u",
    "meanfield",
    "oscillator",
    "burgers_galerkin",
    "mhd",
    "enzyme",
    "yeast",
    "bacterial",
    "pendulum_on_cart",
    "kinematic_commonroad",
    "double_pendulum",
]



import warnings
from itertools import product as iproduct
from typing import Optional

import numpy as np
from sklearn.utils.validation import check_is_fitted

from ..utils import AxesArray
from ..utils import comprehend_axes
from .base import BaseFeatureLibrary
from .base import x_sequence_or_item
from .polynomial_library import PolynomialLibrary
from pysindy.differentiation import FiniteDifference


class PDELibrary(BaseFeatureLibrary):
    """Generate a PDE library with custom functions.

    Parameters
    ----------
    function_library : BaseFeatureLibrary, optional (default
        PolynomialLibrary(degree=3,include_bias=False))
        SINDy library with output features representing library_functions to include
        in the library, in place of library_functions.

    derivative_order : int, optional (default 0)
        Order of derivative to take on each input variable,
        can be arbitrary non-negative integer.

    spatial_grid : np.ndarray, optional (default None)
        The spatial grid for computing derivatives

    temporal_grid : np.ndarray, optional (default None)
        The temporal grid if using SINDy-PI with PDEs.

    include_bias : boolean, optional (default False)
        If True (default), then include a bias column, the feature in which
        all polynomial powers are zero (i.e. a column of ones - acts as an
        intercept term in a linear model).
        This is hard to do with just lambda functions, because if the system
        is not 1D, lambdas will generate duplicates.

    include_interaction : boolean, optional (default True)
        This is a different than the use for the PolynomialLibrary. If true,
        it generates all the mixed derivative terms. If false, the library
        will consist of only pure no-derivative terms and pure derivative
        terms, with no mixed terms.

    implicit_terms : boolean
        Flag to indicate if SINDy-PI (temporal derivatives) is being used
        for the right-hand side of the SINDy fit.

    multiindices : list of integer arrays,  (default None)
        Overrides the derivative_order to customize the included derivative
        orders. Each integer array indicates the order of differentiation
        along the corresponding axis for each derivative term.

    differentiation_method : callable,  (default FiniteDifference)
        Spatial differentiation method.

     diff_kwargs: dictionary,  (default {})
        Keyword options to supply to differtiantion_method.

    Attributes
    ----------
    n_features_in_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. The number of output features
        is the product of the number of library functions and the number of
        input features.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.feature_library import PDELibrary
    """

    def __init__(
        self,
        function_library: Optional[BaseFeatureLibrary] = None,
        derivative_order=0,
        spatial_grid=None,
        temporal_grid=None,
        include_bias=False,
        include_interaction=True,
        implicit_terms=False,
        multiindices=None,
        differentiation_method=FiniteDifference,
        diff_kwargs={},
        is_uniform=None,
        periodic=None,
    ):
        self.function_library = function_library
        self.derivative_order = derivative_order
        self.implicit_terms = implicit_terms
        self.include_bias = include_bias
        self.include_interaction = include_interaction
        self.num_trajectories = 1
        self.differentiation_method = differentiation_method
        self.diff_kwargs = diff_kwargs
        if function_library is None:
            self.function_library = PolynomialLibrary(degree=3, include_bias=False)
        if derivative_order < 0:
            raise ValueError("The derivative order must be >0")

        if is_uniform is not None or periodic is not None:
            # DeprecationWarning are ignored by default...
            warnings.warn(
                "is_uniform and periodic have been deprecated."
                "in favor of differetiation_method and diff_kwargs.",
                UserWarning,
            )

        if (spatial_grid is not None and derivative_order == 0) or (
            spatial_grid is None and derivative_order != 0 and temporal_grid is None
        ):
            raise ValueError(
                "Spatial grid and the derivative order must be "
                "defined at the same time if temporal_grid is not being used."
            )

        if temporal_grid is None and implicit_terms:
            raise ValueError(
                "temporal_grid parameter must be specified if implicit_terms "
                " = True (i.e. if you are using SINDy-PI for PDEs)."
            )
        elif not implicit_terms and temporal_grid is not None:
            raise ValueError(
                "temporal_grid parameter is specified only if implicit_terms "
                " = True (i.e. if you are using SINDy-PI for PDEs)."
            )
        if spatial_grid is not None and spatial_grid.ndim == 1:
            spatial_grid = spatial_grid[:, np.newaxis]

        if temporal_grid is not None and temporal_grid.ndim != 1:
            raise ValueError("temporal_grid parameter must be 1D numpy array.")
        if temporal_grid is not None or spatial_grid is not None:
            if spatial_grid is None:
                spatiotemporal_grid = temporal_grid
                spatial_grid = np.array([])
            elif temporal_grid is None:
                spatiotemporal_grid = spatial_grid[
                    ..., np.newaxis, :
                ]  # append a fake time axis
            else:
                spatiotemporal_grid = np.zeros(
                    (
                        *spatial_grid.shape[:-1],
                        len(temporal_grid),
                        spatial_grid.shape[-1] + 1,
                    )
                )
                for ax in range(spatial_grid.ndim - 1):
                    spatiotemporal_grid[..., ax] = spatial_grid[..., ax][
                        ..., np.newaxis
                    ]
                spatiotemporal_grid[..., -1] = temporal_grid
        else:
            spatiotemporal_grid = np.array([])
            spatial_grid = np.array([])

        self.spatial_grid = spatial_grid

        # list of derivatives
        indices = ()
        if np.array(spatiotemporal_grid).ndim == 1:
            spatiotemporal_grid = np.reshape(
                spatiotemporal_grid, (len(spatiotemporal_grid), 1)
            )

        # if want to include temporal terms -> range(len(dims))
        if self.implicit_terms:
            self.ind_range = spatiotemporal_grid.ndim - 1
        else:
            self.ind_range = spatiotemporal_grid.ndim - 2

        for i in range(self.ind_range):
            indices = indices + (range(derivative_order + 1),)

        if multiindices is None:
            multiindices = []
            for ind in iproduct(*indices):
                current = np.array(ind)
                if np.sum(ind) > 0 and np.sum(ind) <= self.derivative_order:
                    multiindices.append(current)
            multiindices = np.array(multiindices)
        num_derivatives = len(multiindices)

        self.num_derivatives = num_derivatives
        self.multiindices = multiindices
        self.spatiotemporal_grid = AxesArray(
            spatiotemporal_grid, comprehend_axes(spatiotemporal_grid)
        )

    def get_feature_names(self, input_features=None):
        """Return feature names for output features.

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        check_is_fitted(self)
        n_features = self.n_features_in_

        if input_features is None:
            input_features = ["x%d" % i for i in range(n_features)]
        feature_names = []
        lib_names = []

        # Include constant term
        if self.include_bias:
            feature_names.append("1")
        # Include any non-derivative terms
        lib_names = self.function_library.get_feature_names(input_features)
        feature_names = feature_names + lib_names

        def derivative_string(multiindex):
            ret = ""
            for axis in range(self.ind_range):
                if self.implicit_terms and (axis == self.spatiotemporal_grid.ax_time,):
                    str_deriv = "t"
                else:
                    str_deriv = str(axis + 1)
                for i in range(multiindex[axis]):
                    ret = ret + str_deriv
            return ret

        # Include derivative terms
        derivative_feature_names = []
        for k in range(self.num_derivatives):
            for j in range(n_features):
                derivative_feature_names.append(
                    input_features[j] + "_" + derivative_string(self.multiindices[k])
                )
        feature_names = feature_names + derivative_feature_names

        # Include mixed non-derivative + derivative terms
        if (
            self.include_interaction
            and len(lib_names) > 0
            and len(derivative_feature_names) > 0
        ):
            feature_names = (
                feature_names
                + np.char.add(
                    np.array(lib_names).reshape(1, -1),
                    np.array(derivative_feature_names).reshape(-1, 1),
                )
                .reshape(-1)
                .tolist()
            )

        return feature_names

    @x_sequence_or_item
    def fit(self, x_full, y=None):
        """Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            Measurement data.

        Returns
        -------
        self : instance
        """
        x0 = x_full[0]
        n_features = x0.shape[x0.ax_coord]
        self.n_features_in_ = n_features
        n_output_features = 0

        # Count the number of non-derivative terms
        self.function_library.fit(x0.take(0, x0.ax_time))
        n_output_features = self.function_library.n_output_features_

        # Add the mixed derivative library_terms
        if self.include_interaction:
            n_output_features += n_output_features * n_features * self.num_derivatives
        # Add the pure derivative library terms
        n_output_features += n_features * self.num_derivatives

        # If there is a constant term, add 1 to n_output_features
        if self.include_bias:
            n_output_features += 1

        self.n_output_features_ = n_output_features

        # required to generate the function names
        self.get_feature_names()

        return self

    @x_sequence_or_item
    def transform(self, x_full):
        """Transform data to pde features

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray, shape (n_samples, n_output_features)
            The matrix of features, where n_output_features is the number of
            features generated from the tensor product of the derivative terms
            and the library_functions applied to combinations of the inputs.
        """
        check_is_fitted(self)

        xp_full = []
        for x in x_full:
            n_features = x.shape[x.ax_coord]

            if n_features != self.n_features_in_:
                raise ValueError("x shape does not match training shape")

            shape = np.array(x.shape)
            shape[-1] = self.n_output_features_
            xp = np.empty(shape, dtype=x.dtype)

            # derivative terms
            shape[-1] = n_features * self.num_derivatives
            library_derivatives = AxesArray(np.empty(shape, dtype=x.dtype), x.axes)
            library_idx = 0
            for multiindex in self.multiindices:
                derivs = x
                for axis in range(self.ind_range):
                    if multiindex[axis] > 0:
                        s = [0 for dim in self.spatiotemporal_grid.shape]
                        s[axis] = slice(self.spatiotemporal_grid.shape[axis])
                        s[-1] = axis

                        derivs = self.differentiation_method(
                            d=multiindex[axis],
                            axis=axis,
                            **self.diff_kwargs,
                        )._differentiate(derivs, self.spatiotemporal_grid[tuple(s)])
                library_derivatives[
                    ..., library_idx : library_idx + n_features
                ] = derivs
                library_idx += n_features

            # library function terms
            library_functions = self.function_library.fit_transform(x)
            n_library_terms = library_functions.shape[-1]

            library_idx = 0

            # constant term
            if self.include_bias:
                shape[-1] = 1
                xp[..., library_idx] = np.ones(shape[:-1], dtype=x.dtype)
                library_idx += 1

            # library function terms
            xp[..., library_idx : library_idx + n_library_terms] = library_functions
            library_idx += n_library_terms

            # pure derivative terms
            xp[
                ..., library_idx : library_idx + self.num_derivatives * n_features
            ] = library_derivatives
            library_idx += self.num_derivatives * n_features

            # mixed function derivative terms
            shape[-1] = n_library_terms * self.num_derivatives * n_features
            if self.include_interaction:
                xp[
                    ...,
                    library_idx : library_idx
                    + n_library_terms * self.num_derivatives * n_features,
                ] = np.reshape(
                    library_functions[..., "coord", :]
                    * library_derivatives[..., :, "coord"],
                    shape,
                )
                library_idx += n_library_terms * self.num_derivatives * n_features
            xp = AxesArray(xp, comprehend_axes(xp))
            xp_full.append(xp)
        return xp_full

    def get_spatial_grid(self):
        return self.spatial_grid



import warnings
from typing import Union

import numpy as np
from scipy.linalg import cho_factor
from scipy.linalg import cho_solve
from sklearn.exceptions import ConvergenceWarning

from ..utils import capped_simplex_projection
from ..utils import get_prox
from ..utils import get_regularization
from .base import BaseOptimizer


class SR3(BaseOptimizer):
    """
    Sparse relaxed regularized regression.

    Attempts to minimize the objective function

    .. math::

        0.5\\|y-Xw\\|^2_2 + \\lambda R(u)
        + (0.5 / \\nu)\\|w-u\\|^2_2

    where :math:`R(u)` is a regularization function.
    See the following references for more details:

        Zheng, Peng, et al. "A unified framework for sparse relaxed
        regularized regression: SR3." IEEE Access 7 (2018): 1404-1423.

        Champion, K., Zheng, P., Aravkin, A. Y., Brunton, S. L., & Kutz, J. N.
        (2020). A unified sparse optimization framework to learn parsimonious
        physics-informed models from data. IEEE Access, 8, 169259-169271.

    Parameters
    ----------
    reg_weight_lam : float or np.ndarray[float], shape (n_targets, n_features) \
        optional (default 0.005)
        Determines the strength of the regularization. When the
        regularization function R is the l0 norm, the regularization
        is equivalent to performing hard thresholding.
        Use the method calculate_l0_weight to calculate the weight from the threshold.

        When using weighted regularization, this is the array of weights
        for each library function coefficient.
        Each row corresponds to a measurement variable and each column
        to a function from the feature library.
        Recall that SINDy seeks a matrix :math:`\\Xi` such that
        :math:`\\dot{X} \\approx \\Theta(X)\\Xi`.
        ``reg_weight_lam[i, j]`` should specify the weight to be used for the
        (j + 1, i + 1) entry of :math:`\\Xi`. That is to say it should give the
        weight to be used for the (j + 1)st library function in the equation
        for the (i + 1)st measurement variable.

    regularizer : string, optional (default 'L0')
        Regularization function to use. Currently implemented options
        are 'L0' (L0 norm), 'L1' (L1 norm) and 'L2' (L2 norm).
        Note by 'L2 norm' we really mean
        the squared L2 norm, i.e. ridge regression

    relax_coeff_nu : float, optional (default 1)
        Determines the level of relaxation. Decreasing nu encourages
        w and v to be close, whereas increasing nu allows the
        regularized coefficients v to be farther from w.

    tol : float, optional (default 1e-5)
        Tolerance used for determining convergence of the optimization
        algorithm.

    trimming_fraction : float, optional (default 0.0)
        Fraction of the data samples to trim during fitting. Should
        be a float between 0.0 and 1.0. If 0.0, trimming is not
        performed.

    trimming_step_size : float, optional (default 1.0)
        Step size to use in the trimming optimization procedure.

    max_iter : int, optional (default 30)
        Maximum iterations of the optimization algorithm.

    initial_guess : np.ndarray, shape (n_features) or (n_targets, n_features), \
            optional (default None)
        Initial guess for coefficients ``coef_``.
        If None, least-squares is used to obtain an initial guess.

    normalize_columns : boolean, optional (default False)
        Normalize the columns of x (the SINDy library terms) before regression
        by dividing by the L2-norm. Note that the 'normalize' option in sklearn
        is deprecated in sklearn versions >= 1.0 and will be removed.

    copy_X : boolean, optional (default True)
        If True, X will be copied; else, it may be overwritten.

    verbose : bool, optional (default False)
        If True, prints out the different error terms every
        max_iter / 10 iterations.

    unbias: bool (default False)
        See base class for definition.  Most options are incompatible
        with unbiasing.

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Regularized weight vector(s). This is the v in the objective
        function.

    coef_full_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s) that are not subjected to the regularization.
        This is the w in the objective function.

    history_ : list
        History of sparse coefficients. ``history_[k]`` contains the
        sparse coefficients (v in the optimization objective function)
        at iteration k.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.integrate import odeint
    >>> from pysindy import SINDy
    >>> from pysindy.optimizers import SR3
    >>> lorenz = lambda z,t : [10 * (z[1] - z[0]),
    >>>                        z[0] * (28 - z[2]) - z[1],
    >>>                        z[0] * z[1] - 8 / 3 * z[2]]
    >>> t = np.arange(0, 2, .002)
    >>> x = odeint(lorenz, [-8, 8, 27], t)
    >>> opt = SR3(reg_weight_lam=0.1, relax_coeff_nu=1)
    >>> model = SINDy(optimizer=opt)
    >>> model.fit(x, t=t[1] - t[0])
    >>> model.print()
    x0' = -10.004 1 + 10.004 x0
    x1' = 27.994 1 + -0.993 x0 + -1.000 1 x1
    x2' = -2.662 x1 + 1.000 1 x0
    """

    def __init__(
        self,
        reg_weight_lam=0.005,
        regularizer="L0",
        relax_coeff_nu=1.0,
        tol=1e-5,
        trimming_fraction=0.0,
        trimming_step_size=1.0,
        max_iter=30,
        copy_X=True,
        initial_guess=None,
        normalize_columns=False,
        verbose=False,
        unbias=False,
    ):
        super(SR3, self).__init__(
            max_iter=max_iter,
            initial_guess=initial_guess,
            copy_X=copy_X,
            normalize_columns=normalize_columns,
            unbias=unbias,
        )
        if relax_coeff_nu <= 0:
            raise ValueError("relax_coeff_nu must be positive")
        if tol <= 0:
            raise ValueError("tol must be positive")
        if (trimming_fraction < 0) or (trimming_fraction > 1):
            raise ValueError("trimming fraction must be between 0 and 1")
        if trimming_fraction > 0 and unbias:
            raise ValueError(
                "Unbiasing counteracts some of the effects of trimming.  Either set"
                " unbias=False or trimming_fraction=0.0"
            )
        if regularizer.lower() not in (
            "l0",
            "l1",
            "l2",
            "weighted_l0",
            "weighted_l1",
            "weighted_l2",
        ):
            raise NotImplementedError(
                "Please use a valid thresholder, l0, l1, l2, "
                "weighted_l0, weighted_l1, weighted_l2."
            )
        if regularizer[:8].lower() == "weighted" and not isinstance(
            reg_weight_lam, np.ndarray
        ):
            raise ValueError(
                "weighted regularization requires the reg_weight_lam parameter "
                "to be a 2d array"
            )
        if np.any(reg_weight_lam < 0):
            raise ValueError("reg_weight_lam cannot contain negative entries")
        if isinstance(reg_weight_lam, np.ndarray):
            reg_weight_lam = reg_weight_lam.T
        self.reg_weight_lam = reg_weight_lam
        self.relax_coeff_nu = relax_coeff_nu
        self.tol = tol
        self.regularizer = regularizer
        self.reg = get_regularization(regularizer)
        self.prox = get_prox(regularizer)
        if trimming_fraction == 0.0:
            self.use_trimming = False
        else:
            self.use_trimming = True
        self.trimming_fraction = trimming_fraction
        self.trimming_step_size = trimming_step_size
        self.verbose = verbose

    @staticmethod
    def calculate_l0_weight(
        threshold: Union[float, np.ndarray[np.float64]], relax_coeff_nu: float
    ):
        """
        Calculate the L0 regularizer weight that is equivalent to a known L0 threshold

        See Appendix S1 of the following paper for more details.
            Champion, K., Zheng, P., Aravkin, A. Y., Brunton, S. L., & Kutz, J. N.
            (2020). A unified sparse optimization framework to learn parsimonious
            physics-informed models from data. IEEE Access, 8, 169259-169271.
        """
        return (threshold**2) / (2 * relax_coeff_nu)

    def enable_trimming(self, trimming_fraction):
        """
        Enable the trimming of potential outliers.

        Parameters
        ----------
        trimming_fraction: float
            The fraction of samples to be trimmed.
            Must be between 0 and 1.
        """
        self.use_trimming = True
        self.trimming_fraction = trimming_fraction

    def disable_trimming(self):
        """Disable trimming of potential outliers."""
        self.use_trimming = False
        self.trimming_fraction = None

    def _objective(self, x, y, q, coef_full, coef_sparse, trimming_array=None):
        """Objective function"""
        if q != 0:
            print_ind = q % (self.max_iter // 10.0)
        else:
            print_ind = q
        R2 = (y - np.dot(x, coef_full)) ** 2
        D2 = (coef_full - coef_sparse) ** 2
        if self.use_trimming:
            assert trimming_array is not None
            R2 *= trimming_array.reshape(x.shape[0], 1)
        regularization = self.reg(coef_full, self.reg_weight_lam)
        if print_ind == 0 and self.verbose:
            row = [
                q,
                np.sum(R2),
                np.sum(D2) / self.relax_coeff_nu,
                regularization,
                np.sum(R2) + np.sum(D2) + regularization,
            ]
            print(
                "{0:10d} ... {1:10.4e} ... {2:10.4e} ... {3:10.4e}"
                " ... {4:10.4e}".format(*row)
            )
        return (
            0.5 * np.sum(R2)
            + 0.5 * regularization
            + 0.5 * np.sum(D2) / self.relax_coeff_nu
        )

    def _update_full_coef(self, cho, x_transpose_y, coef_sparse):
        """Update the unregularized weight vector"""
        b = x_transpose_y + coef_sparse / self.relax_coeff_nu
        coef_full = cho_solve(cho, b)
        self.iters += 1
        return coef_full

    def _update_sparse_coef(self, coef_full):
        """Update the regularized weight vector"""
        coef_sparse = self.prox(coef_full, self.reg_weight_lam * self.relax_coeff_nu)
        return coef_sparse

    def _update_trimming_array(self, coef_full, trimming_array, trimming_grad):
        trimming_array = trimming_array - self.trimming_step_size * trimming_grad
        trimming_array = capped_simplex_projection(
            trimming_array, self.trimming_fraction
        )
        self.history_trimming_.append(trimming_array)
        return trimming_array

    def _convergence_criterion(self):
        """Calculate the convergence criterion for the optimization"""
        this_coef = self.history_[-1]
        if len(self.history_) > 1:
            last_coef = self.history_[-2]
        else:
            last_coef = np.zeros_like(this_coef)
        err_coef = np.sqrt(np.sum((this_coef - last_coef) ** 2)) / self.relax_coeff_nu
        if self.use_trimming:
            this_trimming_array = self.history_trimming_[-1]
            if len(self.history_trimming_) > 1:
                last_trimming_array = self.history_trimming_[-2]
            else:
                last_trimming_array = np.zeros_like(this_trimming_array)
            err_trimming = (
                np.sqrt(np.sum((this_trimming_array - last_trimming_array) ** 2))
                / self.trimming_step_size
            )
            return err_coef + err_trimming
        return err_coef

    def _reduce(self, x, y):
        """
        Perform at most ``self.max_iter`` iterations of the SR3 algorithm.

        Assumes initial guess for coefficients is stored in ``self.coef_``.
        """
        if self.initial_guess is not None:
            self.coef_ = self.initial_guess

        coef_sparse = self.coef_.T
        n_samples, n_features = x.shape

        coef_full = coef_sparse.copy()
        if self.use_trimming:
            trimming_array = np.repeat(1.0 - self.trimming_fraction, n_samples)
            self.history_trimming_ = [trimming_array]
        else:
            trimming_array = None

        # Precompute some objects for upcoming least-squares solves.
        # Assumes that self.nu is fixed throughout optimization procedure.
        cho = cho_factor(
            np.dot(x.T, x) + np.diag(np.full(x.shape[1], 1.0 / self.relax_coeff_nu))
        )
        x_transpose_y = np.dot(x.T, y)

        # Print initial values for each term in the optimization
        if self.verbose:
            row = [
                "Iteration",
                "|y - Xw|^2",
                "|w-u|^2/v",
                "R(u)",
                "Total Error: |y-Xw|^2 + |w-u|^2/v + R(u)",
            ]
            print(
                "{: >10} ... {: >10} ... {: >10} ... {: >10} ... {: >10}".format(*row)
            )
        objective_history = []

        for k in range(self.max_iter):
            if self.use_trimming:
                x_weighted = x * trimming_array.reshape(n_samples, 1)
                cho = cho_factor(
                    np.dot(x_weighted.T, x)
                    + np.diag(np.full(x.shape[1], 1.0 / self.relax_coeff_nu))
                )
                x_transpose_y = np.dot(x_weighted.T, y)
                trimming_grad = 0.5 * np.sum((y - x.dot(coef_full)) ** 2, axis=1)
            coef_full = self._update_full_coef(cho, x_transpose_y, coef_sparse)
            coef_sparse = self._update_sparse_coef(coef_full)
            self.history_.append(coef_sparse.T)
            if self.use_trimming:
                trimming_array = self._update_trimming_array(
                    coef_full, trimming_array, trimming_grad
                )
            objective_history.append(
                self._objective(x, y, k, coef_full, coef_sparse, trimming_array)
            )
            if self._convergence_criterion() < self.tol:
                # Could not (further) select important features
                break
        else:
            warnings.warn(
                f"SR3 did not converge after {self.max_iter} iterations.",
                ConvergenceWarning,
            )
        self.coef_ = coef_sparse.T
        self.coef_full_ = coef_full.T
        if self.use_trimming:
            self.trimming_array = trimming_array
        self.objective_history = objective_history
